from .group_box import GroupBox
from .labeled_checkbox import LabeledCheckBox
from .line_edit import LineEditComponent
from .radio_button import RadioButton
from .separator import Separator
from .splitter import Splitter
from .tree_view import RioTreeItem, RioTreeView

__all__ = ['GroupBox', 'LabeledCheckBox', 'LineEditComponent', 'RadioButton', 'RioTreeItem', 'RioTreeView', 'Separator', 'Splitter']



if __name__ == '__main__':
    from core_10x.code_samples.person import Person
    from core_10x.exec_control import INTERACTIVE
    from core_10x.ts_store import TsStore

    from ui_10x.collection_editor import Collection, CollectionEditor
    from ui_10x.utils import UxDialog, ux

    ux.init()

    with TsStore.instance_from_uri('mongodb://localhost/test'):
        with INTERACTIVE():
            coll = Collection(cls = Person)
            ce = CollectionEditor(coll = coll)
            w = ce.main_widget()
            d = UxDialog(w)
            d.exec()



from datetime import datetime

from core_10x.attic.backbone.backbone_traitable import RC, T, VaultTraitable
from core_10x.attic.vault import SecKeys


class ResourceAccessor(VaultTraitable):
    # fmt: off
    resource_name: str      = T(T.ID)
    #username: str           = T(T.ID)

    login: str              = T()
    password: bytes         = T()
    last_updated: datetime  = T()
    # fmt: on

    def last_updated_get(self) -> datetime:
        return datetime.utcnow()


class Vault:
    ANY_HOST = '.'

    @staticmethod
    def save_resource_accessor(username: str, hostname: str, login: str, password: str, public_key: bytes = None) -> RC:
        assert login and password, 'login and password must not be empty'

        pwd = SecKeys.encrypt(password, public_key=public_key)
        if pwd is None:
            return RC(False, 'Security Keys are missing. If you are using a new computer, please run "xx user new machine" from your shell')

        ra = ResourceAccessor(hostname=hostname)
        ra.reload()
        rc = ra.set_values(login=login, password=pwd)
        if not rc:
            return rc

        ra.save()



from __future__ import annotations

from core_10x.named_constant import Enum

import ui_10x.platform_interface as i
import ui_10x.rio.components as rio_components
from ui_10x.rio.component_builder import Widget


class Direction(Enum):
    VERTICAL = 'vertical'
    HORIZONTAL = 'horizontal'


Horizontal = Direction.HORIZONTAL


class Splitter(Widget, i.Splitter):
    s_component_class = rio_components.Splitter
    s_pass_children_in_kwargs = True

    def _make_kwargs(self, **kwargs):
        kwargs = super()._make_kwargs(**kwargs)
        if kwargs['direction'] == Horizontal.label:
            del kwargs['align_y']
            kwargs['grow_y'] = True
        kwargs['child_proportions'] = []
        return kwargs

    def __init__(self, direction: Direction = Horizontal, **kwargs):
        super().__init__(**kwargs | dict(direction=direction.label))

    def add_widget(self, widget: Widget):
        self.add_children(widget)
        self['child_proportions'].append(1)

    def set_handle_width(self, width: int):
        self['handle_size'] = width

    def set_stretch_factor(self, widget_index: int, factor: int):
        self['child_proportions'][widget_index] = factor  # TODO: normalize?

    def replace_widget(self, widget_index: int, widget: Widget):
        self.get_children()[widget_index] = widget
        self.force_update()



from numpy import random
from datetime import datetime

from core_10x.traitable import Traitable, T, RT
from core_10x.exec_control import GRAPH_ON
from core_10x.logger import PerfTimer

class Calc(Traitable):
    count: int      = RT(10**4//2)

    seed: int       = RT()
    avg: float      = RT(1.)
    std: float      = RT()
    max_qty: int    = RT(100)

    price: float    = RT()

    def std_get(self):
        return self.avg / 10.

    def price_get(self):
        random.seed(self.seed)
        r = 0.
        for i in range(self.count):
            p = random.normal(self.avg, self.std)
            q = random.randint(1, self.max_qty)
            r += p * q

        return r

if __name__ == '__main__':
    from core_10x.jit.getter_compiler import GetterCompiler

    # graph = GRAPH_ON()
    # graph.begin_using()

    seed = int(datetime.utcnow().timestamp())
    calc = Calc(seed = seed)

    with PerfTimer() as t:
        p1 = calc.price

    dt1 = t.elapsed

    with PerfTimer() as t:
        p2 = calc.price

    dt2 = t.elapsed

    trait_name = 'price'
    gc = GetterCompiler(traitable_class = Calc, trait_name = trait_name)
    trait = Calc.trait(trait_name)
    trait.set_f_get(gc.modified_getter, True)

    p = calc.price

    with PerfTimer() as t:
        p3 = calc.price

    dt3 = t.elapsed

    print(f'prices: {p1:.4f}, {p3:.4f}')
    print(f'acceleration: {dt1/dt3: .1f}')


from core_10x.code_samples.person import Person
from core_10x.exec_control import BTP, GRAPH_OFF, GRAPH_ON
from core_10x.xnone import XNone


def test():
    p = Person(first_name = 'Sasha', last_name = 'Davidovich')
    #p.weight_lbs = 100
    #assert p.weight == 100

    p.invalidate_value(p.T.weight)
    p.invalidate_value(p.T.weight_lbs)

    assert p.weight is XNone

if __name__ == '__main__':

    with GRAPH_ON(convert_values = True, debug = True):
        print(BTP.current().flags())
        test()
        with GRAPH_OFF():
            test()

    p = Person(first_name = 'Sasha', last_name = 'Davidovich')
    with GRAPH_ON(convert_values = True):
        p.weight_lbs = 'a'





import sys

from core_10x.traitable import RC, RC_TRUE, Traitable


class TraitableCli(Traitable):
    s_master = None
    s_switch = {}

    def __init_subclass__(cls, _command: str = None, **kwargs):
        cls.s_switch = {}
        if cls.s_master is None:  # -- master parser
            assert _command is None, 'master may not have a command'
            cls.s_master = cls
        else:  # -- subordinate parser
            assert _command and _command.isidentifier(), 'command must be a valid identifier'
            cls.s_master.s_switch[_command] = cls
            cls.s_master = cls

        super().__init_subclass__(**kwargs)

    @classmethod
    def from_command_line(cls) -> tuple:  # -- (RC, Traitable)
        """
        Creates an instance of target traitable parsing positional args to obtain the right target class and then taking trait values
        in the name = value pairs form.
        One can use any spacing around symbol '=', i.e.: name=value name= value name =value or name = value

        :return:    (rc, traitable) - RC and a traitable according to args and trait values parsed from sys.argv.
                    rc holds error message(s) if instantiation fails for any reason (traitable is set to None)
        """
        _script, *args = sys.argv
        return cls.instance_from_args(args)

    @classmethod
    def instance_from_args(cls, input_args: tuple) -> tuple:
        args = []
        trait_values = {}
        rc = cls.parse(input_args, args, trait_values)
        if not rc:
            return rc, None

        return cls.instantiate(args, trait_values)

    @classmethod
    def instantiate(cls, args, trait_values: dict) -> tuple:  # -- (RC, target_traitable)
        if not args:
            try:
                res = cls()
                for name, value in trait_values.items():
                    trait = cls.trait(name)
                    if not trait:
                        return RC(False, f'unknown attribute {name}\nValid attributes: {", ".join(cls.s_dir)}'), None

                    rc = res.set_value(trait, value)
                    if not rc:
                        return RC(False, f'{name}: {rc.err()}'), None

                return RC_TRUE, res

            except Exception as ex:
                return RC(False, str(ex)), None

        command, *args = args
        parser = cls.s_switch.get(command)
        if not parser or not issubclass(parser, TraitableCli):
            return RC(False, f'Unknown argument {command}'), None

        return parser.instantiate(args, trait_values)

    @classmethod
    def parse(cls, input_args: tuple, args: list, trait_values: dict) -> RC:  # noqa: C901
        if not input_args:
            return RC_TRUE

        trait_name = None
        deal_with_args = True
        new_vp = True
        eq_expected = False
        for i, arg in enumerate(input_args):
            if deal_with_args:
                parts = arg.split('=', 1)
                n = len(parts)
                if n == 1:
                    args.append(arg)
                else:
                    deal_with_args = False
                    eq_expected = False
                    first, second = parts
                    if not first and not second:  # -- just '='
                        if i:
                            trait_name = args.pop(-1)
                            new_vp = False
                        else:
                            return RC(False, 'May not start with "="')

                    elif first:
                        if not second:  # -- xxx=
                            trait_name = first
                            new_vp = False
                        else:  # -- xxx=yyy
                            trait_values[first] = second
                            new_vp = True
                    else:  # -- =yyy
                        if i:
                            trait_name = args.pop(-1)
                            trait_values[trait_name] = second
                            new_vp = True
                        else:
                            return RC(False, 'May not start with "="')

            else:
                parts = arg.split('=', 1)
                n = len(parts)
                if new_vp:
                    trait_name = parts[0]
                    if n == 1:
                        eq_expected = True
                        new_vp = False
                    else:
                        value = parts[1]
                        if value:
                            trait_values[trait_name] = value
                            new_vp = True
                        else:
                            eq_expected = False
                            new_vp = False
                else:
                    if not eq_expected:
                        if n == 2:
                            return RC(False, f'Invalid "=": {input_args[i - 1]} {arg}')

                        trait_values[trait_name] = parts[0]
                        new_vp = True

                    else:
                        if n == 1:
                            return RC(False, f'"=" is expected before {arg}')

                        if parts[0]:
                            return RC(False, f'{input_args[i - 1]} {arg}')

                        if parts[1]:
                            trait_values[trait_name] = parts[1]
                            new_vp = True

                        else:
                            eq_expected = False

        if not new_vp:
            return RC(False, f'Value is missing for {trait_name} = ')

        return RC_TRUE



from datetime import datetime
import time

import pytest

from core_10x.testlib.fixtures import with_transactions
from core_10x.testlib.traitable_history_tests import TestTraitableHistory, test_collection, test_store

@pytest.fixture
def clock_freezer():
    class Freezer:
        @staticmethod
        def utcnow():
            time.sleep(0.001)
            return datetime.utcnow()
    return Freezer



import asyncio

import rio.testing.browser_client
from ui_10x.rio.component_builder import DynamicComponent
from ui_10x.rio.widgets import PushButton


async def test_button_comprehensive() -> None:
    """Test PushButton with comprehensive client-widget interaction verification."""
    widget = PushButton('Click Me')
    find_button_text = 'document.querySelector(".rio-button .rio-text").children[0].innerText'

    clicked_calls = []

    def clicked_handler():
        clicked_calls.append(True)

    widget.clicked_connect(clicked_handler)

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # 1) Verify client shows widget value
        assert widget['content'] == 'Click Me'
        client_text = await test_client.execute_js(find_button_text)
        assert client_text == 'Click Me'

        # 2) Modify client value (user clicking button)
        await test_client.click(10, 10)  # Click on the button

        # 3) Verify widget reflects client value (click handler called)
        assert len(clicked_calls) == 1
        assert clicked_calls[0] is True

        # Test another click
        await test_client.click(10, 10)
        assert len(clicked_calls) == 2


async def test_button_disabled_interaction() -> None:
    """Test PushButton disabled state blocks user interaction."""
    widget = PushButton('Test Button')

    clicked_calls = []

    def clicked_handler():
        clicked_calls.append(True)

    widget.clicked_connect(clicked_handler)

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial enabled state - clicks should work
        await test_client.click(10, 10)
        assert len(clicked_calls) == 1

        # Disable widget
        widget.set_enabled(False)
        await test_client.wait_for_refresh()
        assert not widget['is_sensitive']

        # Test that clicks are blocked when disabled
        await test_client.click(10, 10)
        assert len(clicked_calls) == 1  # No additional calls

        # Re-enable widget
        widget.set_enabled(True)
        await test_client.wait_for_refresh()
        assert widget['is_sensitive']

        # Test that clicks work again
        await test_client.click(10, 10)
        assert len(clicked_calls) == 2


async def test_button_basic_functionality() -> None:
    """Test PushButton basic functionality."""
    widget = PushButton('Test Button')

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial state
        assert widget['content'] == 'Test Button'
        assert widget['is_sensitive']

        # Test enabled/disabled state with timeout protection
        widget.set_enabled(False)
        await test_client.wait_for_refresh()
        assert not widget['is_sensitive']

        widget.set_enabled(True)
        await test_client.wait_for_refresh()
        assert widget['is_sensitive']

        # Test flat style changes with timeout protection
        assert not await test_client.execute_js('document.querySelector(".rio-buttonstyle-plain-text")')
        widget.set_flat(True)
        await test_client.wait_for_refresh()
        assert await test_client.execute_js('document.querySelector(".rio-buttonstyle-plain-text")')



from __future__ import annotations

import ui_10x.platform_interface as i
import ui_10x.rio.components as rio_components
from ui_10x.rio.component_builder import Widget


class GroupBox(Widget, i.GroupBox):
    s_component_class = rio_components.GroupBox
    s_pass_children_in_kwargs = True
    s_unwrap_single_child = False

    def __init__(self, *args, **kwargs):
        _parent = None
        title = kwargs.pop('title', '')
        children = ()
        if len(args) == 1:
            _parent = args[0]
        elif len(args) >= 2:
            assert not title, 'title specified twice'
            _parent, title = args[:2]
            children = args[2:]
        super().__init__(*children, **kwargs)
        self['title'] = title

    def set_title(self, title: str):
        self['title'] = title



if __name__ == '__main__':

    from datetime import date

    from xx_common.xxcalendar import Calendar, CalendarAdjustment

    us_c = Calendar(_replace = True,
        name                = 'US',
        non_working_days    = [
            date(2025,1,1), date(2025,1,20), date(2025,2,17), date(2025,5,26), date(2025,7,4),
            date(2025,9,1), date(2025,10,13), date(2025,11,11), date(2025,11,27), date(2025,12, 25)
        ]
    )

    uk_c = Calendar(_replace = True,
        name                = 'UK',
        non_working_days    = [
            date(2025,1,1), date(2025,4,18), date(2025,5,5), date(2025,5,26), date(2025,8,25),
            date(2025,12, 25), date(2025,12, 26)
        ]
    )

    c = Calendar.union('US', 'UK')
    union_nwds = c.non_working_days

    c2 = Calendar.intersection('US', 'UK')
    cross_nwds = c2.non_working_days

    sofr_adjust = CalendarAdjustment(_replace = True,
        name        = 'SOFR',
        add_days    = [date(2019, 4, 4)]
    )

    ca = Calendar(name = 'US', adjusted_for = 'SOFR')
    nwds_ca = ca.non_working_days



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import ui_10x.platform_interface as i
import ui_10x.rio.components as rio_components
from ui_10x.rio.component_builder import Widget

# noinspection PyCompatibility
from . import (
    Application,
    Dialog,
    HBoxLayout,
    Label,
    PushButton,
    Style,
    VBoxLayout,
)

if TYPE_CHECKING:
    from collections.abc import Callable


class Separator(Widget, i.Separator):
    s_component_class = rio_components.Separator
    s_forced_kwargs = {}


class MessageBox(i.MessageBox):
    @classmethod
    def _dialog(cls, parent: Widget, title: str, message: str, icon: Style.StandardPixmap, on_close: Callable[[Any], None]) -> bool | None:
        style = Application.style()
        if icon == Style.StandardPixmap.SP_MESSAGEBOXQUESTION:
            buttons = [
                PushButton(style.standard_icon(Style.StandardPixmap.SP_DIALOGYESBUTTON.value), 'Yes', on_press=lambda: dlg.on_accept()),
                PushButton(style.standard_icon(Style.StandardPixmap.SP_DIALOGNOBUTTON.value), 'No', on_press=lambda: dlg.on_reject()),
            ]
        else:
            buttons = [PushButton(style.standard_icon(Style.StandardPixmap.SP_DIALOGOKBUTTON.value), 'Ok', on_press=lambda: dlg.on_accept())]
        children = (
            Label(title, align_x=0.5, align_y=0, style='heading1'),
            Separator(),
            HBoxLayout(
                PushButton(style.standard_icon(icon.value), '', style='plain-text', is_sensitive=False, align_y=0, align_x=0, grow_x=False),
                Label(message, align_y=0, overflow='wrap', grow_x=True),
            ),
            Separator(),
            HBoxLayout(*buttons),
        )
        dlg = Dialog(
            parent=parent, title=title, children=children, on_accept=lambda: on_close(dlg.accepted), on_reject=lambda: on_close(dlg.accepted)
        )
        dlg.set_layout(VBoxLayout())
        return dlg.show() if on_close else dlg.exec()

    @classmethod
    def question(cls, parent: Widget, title: str, message: str, on_close: Callable[[Any], None]) -> bool:
        return cls._dialog(parent=parent, title=title, message=message, icon=Style.StandardPixmap.SP_MESSAGEBOXQUESTION, on_close=on_close)

    @classmethod
    def warning(cls, parent: Widget, title: str, message: str, on_close: Callable[[Any], None]):
        return cls._dialog(parent=parent, title=title, message=message, icon=Style.StandardPixmap.SP_MESSAGEBOXWARNING, on_close=on_close)

    @classmethod
    def information(cls, parent: Widget, title: str, message: str, on_close: Callable[[Any], None]):
        return cls._dialog(parent=parent, title=title, message=message, icon=Style.StandardPixmap.SP_MESSAGEBOXINFORMATION, on_close=on_close)

    @classmethod
    def is_yes_button(cls, sb) -> bool:
        return sb



from __future__ import annotations

import rio
import ui_10x.platform_interface as i
from ui_10x.rio.component_builder import Layout


class BoxLayout(Layout): ...


class VBoxLayout(BoxLayout, i.VBoxLayout):
    s_component_class = rio.Column
    s_stretch_arg = 'grow_y'

    def _make_kwargs(self, **kwargs):
        kwargs = super()._make_kwargs(**kwargs)
        del kwargs['align_y']
        return kwargs


class HBoxLayout(BoxLayout, i.HBoxLayout):
    s_component_class = rio.Row
    s_stretch_arg = 'grow_x'


class FormLayout(Layout, i.FormLayout):
    s_component_class = rio.Grid
    s_children_attr = 'rows'

    def add_row(self, *args):
        self.add_children(args)

    def _build_children(self, session: rio.Session):
        return [[child() for child in children] for children in self._get_children()]  # children are 2d array



from __future__ import annotations

import heapq
import itertools
import operator
from itertools import zip_longest
from typing import TYPE_CHECKING

from core_10x.nucleus import Nucleus
from core_10x.trait_filter import EQ, f
from core_10x.ts_store import TsCollection, TsStore

if TYPE_CHECKING:
    from collections.abc import Iterable


class _OrderKey:
    __slots__ = ('reverse', 'value')

    @classmethod
    def _dict_cmp(cls, d: dict, od: dict) -> int:
        assert isinstance(od, dict), 'can only compare dict to dict'

        for i, oi in zip_longest(d.items(), od.items()):
            if i == oi:
                continue
            if i is None:
                return -1  # all match, d is shorter
            if oi is None:
                return 1  # all match d is longer
            k, v = i
            ok, ov = oi
            if k != ok:
                return -1 if k < ok else 1 if k > ok else 0  # keys do not match
            if isinstance(v, dict) and isinstance(ov, dict):
                return cls._dict_cmp(v, ov)
            # TODO: handle mismatching types..
            return -1 if v < ov else 1 if v > ov else 0
        return 0  # equals..

    def __init__(self, value, reverse: bool):
        self.value = value
        self.reverse = reverse

    def __lt__(self, other):
        assert isinstance(other, _OrderKey), 'can only compare to order key'
        v = self.value
        ov = other.value
        if isinstance(v, dict):
            return self._dict_cmp(v, ov) < 0 if self.reverse else self._dict_cmp(ov, v) > 0
        return ov < v if self.reverse else v < ov

    def __eq__(self, other):
        assert isinstance(other, _OrderKey), 'can only compare to order key'
        v = self.value
        ov = other.value
        return not self._dict_cmp(v, ov) if isinstance(v, dict) else v == ov

    @classmethod
    def key(cls, item, order):
        return tuple(cls(value=item.get(k), reverse=v < 0) for k, v in order)


class TsUnionCollection(TsCollection):
    def __init__(self, *collections: TsCollection):
        self.collections = collections

    def collection_name(self) -> str:
        return self.collections[0].collection_name() if self.collections else ''

    def find(self, query: f = None, _at_most: int = 0, _order: dict = None) -> Iterable:
        order = tuple((_order or {'_id': 1}).items())  # FIX: assumes _id is always there!
        order_key = _OrderKey.key
        iterables = (collection.find(query, _at_most=_at_most, _order=_order) for collection in self.collections)
        keyed_iterables = (((order_key(item, order), item) for item in iterable) for iterable in iterables if iterable is not None)
        results = (item for _, item in heapq.merge(*keyed_iterables, key=operator.itemgetter(0)))
        return (item for i, item in enumerate(results) if i < _at_most) if _at_most else results

    def save_new(self, serialized_traitable: dict, overwrite: bool = False) -> int:
        return self.collections[0].save_new(serialized_traitable, overwrite=overwrite)

    def save(self, serialized_traitable):
        # if serialized_traitable was not loaded from the union head, we need to call save_new
        id_tag = Nucleus.ID_TAG()
        if not self.collections[0].exists(f(**{id_tag: EQ(serialized_traitable[id_tag])})):
            # TODO: optimize by introducing save with overwrite flag to bypass the "inappropriate restore from deleted" check
            return self.collections[0].save_new(serialized_traitable)
        return self.collections[0].save(serialized_traitable)

    def delete(self, id_value):
        # if id_value exists in the union tail, return False as the object wasn't fully deleted
        return self.collections[0].delete(id_value) and not self.exists(f(**{Nucleus.ID_TAG(): EQ(id_value)}))

    def create_index(self, name: str, trait_name: str | list[tuple[str, int]], **index_args) -> str:
        # TODO: verify that index exists in the union tail
        return self.collections[0].create_index(name, trait_name, **index_args)

    def max(self, trait_name: str, filter: f = None) -> dict:
        results = (result for result in (collection.max(trait_name, filter) for collection in self.collections) if result is not None)
        return max(results, key=operator.itemgetter(trait_name), default=None)

    def min(self, trait_name: str, filter: f = None) -> dict:
        results = (result for result in (collection.min(trait_name, filter) for collection in self.collections) if result is not None)
        return min(results, key=operator.itemgetter(trait_name), default=None)

    def id_exists(self, id_value: str) -> bool:
        return any(collection.id_exists(id_value) for collection in self.collections)

    def count(self, query: f = None) -> int:
        return sum(collection.count(query) for collection in self.collections)

    def load(self, id_value: str) -> dict | None:
        for collection in self.collections:
            data = collection.load(id_value)
            if data is not None:
                return data
        return None


class TsUnion(TsStore, resource_name='TS_UNION'):

    @classmethod
    def parse_uri(cls, uri: str) -> dict:
        raise NotImplementedError('TsUnion does not support URI parsing')

    @classmethod
    def is_running_with_auth(cls, host_name: str) -> tuple:  # -- (is_running, with_auth)
        raise NotImplementedError

    @classmethod
    def standard_key(cls, *args) -> tuple:
        return tuple(cls.s_resource_type.resource_driver(kw['driver_name']).standard_key(**kw) for kw in args)

    @classmethod
    def new_instance(cls, *args, **kwargs) -> TsUnion:
        stores = [cls.s_resource_type.resource_driver(kw.pop('driver_name')).instance(**kw) for kw in args]
        return TsUnion(*stores)

    def __init__(self, *stores: TsStore):
        super().__init__()
        self.stores = stores

    def collection_names(self, regexp: str = None) -> list:
        return list(set(itertools.chain(*(store.collection_names(regexp) for store in self.stores))))

    def collection(self, collection_name):
        return TsUnionCollection(*(store.collection(collection_name) for store in self.stores))

    def delete_collection(self, collection_name: str) -> bool:
        return self.stores[0].delete_collection(collection_name) if self.stores else False

    def auth_user(self) -> str | None:
        return self.stores[0].auth_user() if self.stores else None

    def transaction(self):
        """Context manager for transactional operations. Delegates to base (no stores) or first store."""
        return next(iter(self.stores), super()).transaction()



if __name__ == '__main__':
    from core_10x.code_samples.directories import ANIMALS

    from ui_10x.choice import MultiChoice
    from ui_10x.utils import UxDialog, ux

    ux.init()

    choice1 = MultiChoice(choices = ANIMALS)

    d = UxDialog(choice1.widget())
    d.exec()

    choice2 = MultiChoice(choices = ANIMALS.choices())

    d = UxDialog(choice2.widget())
    d.exec()



from datetime import date, datetime

from core_10x.code_samples.person import Person
from core_10x.traitable import T, RT, Traitable, Trait, BTraitFlags


class Event(Traitable):
    at: datetime = T()

class Dummy(Traitable):
    name: str       = T(T.ID)
    payload: float  = T(3.1415)

    view: str       = RT()

    def name_get(self):
        return 'AMD'

    def view_get(self) -> str:
        print(f'{self.__class__}.view called')
        return f'{self.name}: {self.payload}'

from core_10x.named_constant import NamedConstant

class F_KEY(NamedConstant):
    CLASS       = lambda obj: obj.__class__
    MKT_NAME    = lambda obj: obj.mkt_name

class Bucket(Traitable):
    f_key: NamedConstant = T()

if __name__ == '__main__':
    from core_10x.manual_tests.traitable_trivial_test import Dummy, Event, F_KEY, Bucket

    e = Event()
    print(e.id_value())
    rc = e.set_values(at = datetime.now())

    p = Person(first_name = 'Sasha', last_name = 'Davidovich')
    #print(p.id(), p.full_name)

    # p.full_name = 'Ilya Pevzner'
    # print(p.full_name)
    # print(f'first name = {p.first_name}; last name = {p.last_name}')

    p.dob = date(1963, 5, 14)
    #p.age = 61

    #print(f'older_than(25) = {p.older_than(25)}')

    #try:
    #    p.older_than = False
    #except TypeError as ex:
    #    print(ex)

    #p.older_than = trait_value(False, 25)   #-- set with args!
    #print(f'older_than(25) = {p.older_than(25)}')

    #print(p.from_any(p.T.older_than(), 'False'))

    p2 = Person(first_name = 'Sasha', last_name = 'Davidovich')
    #assert p2.age == 61
    print(p2.age)

    #print(Nucleus.serialize_any(p,False))

    name = 'AMD'
    d = Dummy()
    assert d.name == name
    d.payload = 100.

    d2 = Dummy(name = name)
    assert d2.name == d.name

    dx = Dummy.existing_instance(name = name)
    assert dx.payload == 100.

    dx2 = Dummy.existing_instance()
    assert dx2.payload == 100.

    fn_t    = Person.trait('first_name')
    wqu_t   = Person.trait('weight_qu')
    w_t     = Person.trait('weight')

    print(f'first_name = {fn_t.has_custom_getter()}; weight_qu = {wqu_t.has_custom_getter()}; weight = {w_t.has_custom_getter()}')

    p3 = Person(_replace = True, first_name = 'S2', last_name = 'Davidovich2', dob = date(1963, 1, 1), weight_lbs = 250.)
    rc = p3.verify()
    if not rc:
        print(rc.error())

    p4 = Person(_replace = True, first_name = 'S', last_name = 'Davidovich', dob = date(1963, 1, 1), weight_lbs = 250.)
    rc = p4.verify()
    if not rc:
        print(rc.error())

    v = d.view
    v2 = d.view, d.view
    d.payload = -1000
    v2 = d.view


    b = Bucket(f_key = F_KEY.CLASS)
    d = b.serialize_object()


from __future__ import annotations

from py10x_kernel import OsUser

from core_10x.traitable import RT, T
from core_10x.attic.backbone.backbone_traitable import BackboneTraitable
from infra_10x.namespace import FUNCTIONAL_ACCOUNT_PREFIX, USER_ADMIN_SUFFIX
from core_10x.global_cache import cache
from core_10x.trait_filter import f


class BackboneUserGroup(BackboneTraitable):
    # fmt: off
    name: str           = T(T.ID)
    description: str    = T()
    everyone: bool      = T(False)
    is_admin: bool      = T(False)
    is_reserved: bool   = T(False)
    builtin_roles: list = T()
    user_ids: list      = T()

    users: set          = RT()
    # fmt: on

    def users_get(self) -> set:
        return set(self.user_ids)

    def is_member(self, user_id: str) -> bool:
        return user_id in self.users if not self.everyone else False


class BackboneUser(BackboneTraitable):
    # fmt: off
    user_id: str        = T(T.ID)   // 'OS login'
    suspended: bool     = T(False)
    public_key: bytes   = T()

    is_admin: bool      = RT()
    downloaded: bool    = RT(False)
    builtin_roles: list = RT()
    # fmt: on

    def user_id_get(self) -> str:
        return OsUser.me.name()

    def is_admin_get(self) -> bool:
        for group in BackboneUserGroup.collection().find(f(is_admin=True)):
            if group.is_member(self.user_id):
                return True

        return False

    def admin_id(self) -> str:
        return f'{self.user_id}_{USER_ADMIN_SUFFIX}' if self.is_admin else ''

    @classmethod
    def regular_id(cls, admin_id: str) -> str:
        return admin_id.split(f'_{USER_ADMIN_SUFFIX}')[0]

    @classmethod
    def is_functional_account(cls, user_id: str) -> bool:
        return user_id.split('-', 1) == FUNCTIONAL_ACCOUNT_PREFIX

    @classmethod
    @cache
    def me(cls) -> BackboneUser:
        return cls.existing_instance(user_id=OsUser.me.name())



from core_10x.named_constant import Enum, NamedConstantTable


# ruff: noqa: E741
class R:
    # fmt: off
    class Symbol(Enum):
        I = () 
        V = ()
        X = ()
        L = ()
        C = ()
        D = ()
        M = ()

    class Attr(Enum):
        VALUE           = ()
        REPEAT          = ()
        SUBTRACTABLE    = ()

    SYMBOL_TABLE = NamedConstantTable(Symbol, Attr,
        #       VALUE       REPEAT      SUBTRACTABLE
        I = (   1,          3,          True ),
        V = (   5,          1,          False ),
        X = (   10,         3,          True ),
        L = (   50,         1,          False ),
        C = (   100,        3,          True ),
        D = (   500,        1,          False ),
        M = (   1000,       3,          True ),
    )
    # fmt: on
    def __init__(self, text: str):
        text = text.upper()
        x = 0
        prev_values = [0, 0, 0]
        for i, symbol in enumerate(reversed(text)):
            ref = R.SYMBOL_TABLE[symbol]
            v = ref.VALUE

            # -- Rules
            repeat = ref.REPEAT
            over_repeated = all(v == prev_values[i] for i in range(repeat))
            assert not over_repeated, f"'{symbol}' is repeated more than {repeat} times"

            subtractable = ref.SUBTRACTABLE
            if not subtractable and v < prev_values[0]:
                raise AssertionError(f"'{symbol}' may not occur before '{text[-i]}'")

            if v >= prev_values[0]:
                x += v
            else:
                x -= v

            prev_values[2] = prev_values[1]
            prev_values[1] = prev_values[0]
            prev_values[0] = v

        self.roman = text
        self.value = x

    def __repr__(self):
        return f'{self.roman} = {self.value}'


if __name__ == '__main__':
    x = R('IX')
    y = R('MCMLXXXIV')



import inspect

from core_10x.trait_definition import T


class X:
    user_id: str    = T(T.ID | T.RUNTIME)   #, Ui('User ID', tip = 'OS login', min_width = 50))
    age: int        = T()
    #model: any      = RT()


annotations = inspect.get_annotations(X)

#dir = TraitDefinition.dir( X.__dict__, annotations, {} )

#tdef: TraitDefinition = dir['user_id']

# t = Trait(tdef)
# t2 = copy.deepcopy(t)


import asyncio
from datetime import date

import dateutil.parser
import rio.testing
from ui_10x.rio.component_builder import DynamicComponent
from ui_10x.rio.widgets import CalendarWidget

find_selected_date = (
    'document.querySelector(".rio-calendar-selected-day").textContent + " " + document.querySelector(".rio-calendar-year-month-display").textContent'
)


async def verify_content(widget, test_client):
    selected_date = await test_client.execute_js(find_selected_date)

    assert dateutil.parser.parse(selected_date).date() == widget.selected_date()


async def test_calendar_comprehensive() -> None:
    """Test CalendarWidget with comprehensive client-widget interaction verification."""
    widget = CalendarWidget()
    test_date = date(2024, 6, 15)

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # 1) Verify client shows widget value (initial state)
        assert widget.selected_date() == date.today()

        # Verify client shows the expected date
        await verify_content(widget, test_client)

        # 2) Modify widget value and verify client reflects it
        widget.set_selected_date(test_date)
        await test_client.wait_for_refresh()
        assert widget.selected_date() == test_date

        # Verify client shows the expected date
        await verify_content(widget, test_client)

        # Test another date change
        new_date = date(2024, 8, 20)
        widget.set_selected_date(new_date)
        await test_client.wait_for_refresh()
        assert widget.selected_date() == new_date

        # Verify client shows the new expected date
        await verify_content(widget, test_client)


async def test_calendar_disabled_interaction() -> None:
    """Test CalendarWidget disabled state."""
    widget = CalendarWidget()

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial enabled state
        assert widget['is_sensitive']

        # Disable widget
        widget.set_enabled(False)
        await test_client.wait_for_refresh()
        assert not widget['is_sensitive']
        await verify_content(widget, test_client)

        # Test that date changes still work (programmatic changes are allowed)
        widget.set_selected_date(date(2024, 12, 25))
        await test_client.wait_for_refresh()
        assert widget.selected_date() == date(2024, 12, 25)

        # Verify client shows the expected date even when disabled
        await verify_content(widget, test_client)

        # Re-enable widget
        widget.set_enabled(True)
        await test_client.wait_for_refresh()
        assert widget['is_sensitive']

        # Test that date changes continue to work
        widget.set_selected_date(date(2024, 1, 1))
        await test_client.wait_for_refresh()
        assert widget.selected_date() == date(2024, 1, 1)

        # Verify client shows the final expected date
        await verify_content(widget, test_client)


async def test_calendar_basic_functionality() -> None:
    """Test CalendarWidget basic functionality."""
    widget = CalendarWidget()

    async with rio.testing.DummyClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial state
        assert widget is not None
        assert widget['is_sensitive']

        # Test date setting
        test_date = date(2024, 3, 10)
        widget.set_selected_date(test_date)
        await test_client.wait_for_refresh()
        assert widget.selected_date() == test_date

        # Test enabled/disabled state
        widget.set_enabled(False)
        await test_client.wait_for_refresh()
        assert not widget['is_sensitive']

        widget.set_enabled(True)
        await test_client.wait_for_refresh()
        assert widget['is_sensitive']



from __future__ import annotations

from datetime import date

import pytest
from core_10x.exec_control import CACHE_ONLY
from xx_common.xxcalendar import Calendar, CalendarAdjustment, CalendarNameParser


class TestCalendarNameParser:
    """Unit tests for CalendarNameParser class."""

    def test_combo_name(self):
        """Test combo_name method."""
        assert CalendarNameParser.combo_name('CAL1|CAL2') is True
        assert CalendarNameParser.combo_name('CAL1&CAL2') is True
        assert CalendarNameParser.combo_name('SIMPLE_CAL') is False

    def test_operation_repr_with_calendar_objects(self):
        """Test operation_repr with calendar objects."""

        # Create mock calendar objects for testing
        class MockCalendar:
            def __init__(self, name):
                self.name = name

        cal1 = MockCalendar('CAL1')
        cal2 = MockCalendar('CAL2')
        cal3 = MockCalendar('CAL3')

        result = CalendarNameParser.operation_repr(MockCalendar, CalendarNameParser.OR_CHAR, cal1, cal2, cal3)
        # The result includes a trailing comma before the operation and a newline at the end
        expected = 'CAL1,CAL2,CAL3,|3\n'
        assert result == expected

        result = CalendarNameParser.operation_repr(MockCalendar, CalendarNameParser.AND_CHAR, cal1, cal2)
        expected = 'CAL1,CAL2,&2\n'
        assert result == expected

    def test_operation_repr_with_strings(self):
        """Test operation_repr with string names."""
        # This would require existing calendars in the database
        # We'll test the assertion failures instead
        pass

    def test_operation_repr_invalid_op(self):
        """Test operation_repr with invalid operation character."""

        class MockCalendar:
            def __init__(self, name):
                self.name = name

        cal1 = MockCalendar('CAL1')
        cal2 = MockCalendar('CAL2')

        with pytest.raises(AssertionError, match='Unknown op char = \\*'):
            CalendarNameParser.operation_repr(MockCalendar, '*', cal1, cal2)

    def test_operation_repr_too_few_calendars(self):
        """Test operation_repr with too few calendars."""

        class MockCalendar:
            def __init__(self, name):
                self.name = name

        cal1 = MockCalendar('CAL1')

        with pytest.raises(AssertionError, match='there must be at least 2 calendars'):
            CalendarNameParser.operation_repr(MockCalendar, CalendarNameParser.OR_CHAR, cal1)

    def test_parse_simple_calendar(self, ts_instance):
        """Test parse method with simple calendar name."""
        from datetime import date

        with ts_instance:
            holidays = {date(2024, 1, 1), date(2024, 12, 25)}
            cal = Calendar(
                name='TEST_SIMPLE_CAL',
                description='Simple calendar for parsing',
                non_working_days=sorted(holidays),
                _replace=True,
            )
            assert cal.save()

            parsed_days = CalendarNameParser.parse(Calendar, cal.name)
            assert parsed_days == holidays

    def test_parse_combined_calendar(self, ts_instance):
        """Test parse method with combined calendar operations."""
        from datetime import date

        with ts_instance:
            holidays_a = {date(2024, 1, 1), date(2024, 1, 2)}
            holidays_b = {date(2024, 1, 2), date(2024, 1, 3)}

            cal_a = Calendar(
                name='TEST_CAL_A',
                description='Calendar A for combined parsing',
                non_working_days=sorted(holidays_a),
                _replace=True,
            )
            cal_b = Calendar(
                name='TEST_CAL_B',
                description='Calendar B for combined parsing',
                non_working_days=sorted(holidays_b),
                _replace=True,
            )

            assert cal_a.save()
            assert cal_b.save()

            # OR operation: union of non-working days
            union_name = CalendarNameParser.operation_repr(Calendar, CalendarNameParser.OR_CHAR, cal_a, cal_b)
            union_days = CalendarNameParser.parse(Calendar, union_name)
            assert union_days == holidays_a | holidays_b

            # AND operation: intersection of non-working days
            intersection_name = CalendarNameParser.operation_repr(Calendar, CalendarNameParser.AND_CHAR, cal_a, cal_b)
            intersection_days = CalendarNameParser.parse(Calendar, intersection_name)
            assert intersection_days == holidays_a & holidays_b


class TestCalendarAdjustment:
    """Unit tests for CalendarAdjustment class."""

    def test_creation(self):
        """Test CalendarAdjustment can be created."""
        with CACHE_ONLY():
            # CalendarAdjustment requires name (which has T.ID flag)
            ca = CalendarAdjustment(name='TEST_ADJ', _replace=True)
            assert ca.name == 'TEST_ADJ'

    def test_calendar_adjustment_is_storable(self):
        """Test that CalendarAdjustment is storable."""
        assert CalendarAdjustment.is_storable()
        assert CalendarAdjustment.trait('name')  # Has ID trait

    def test_calendar_adjustment_with_adjusted_for(self, ts_instance):
        """Test Calendar with adjusted_for parameter applies CalendarAdjustment."""
        with ts_instance:
            # Create base calendar
            base_cal = Calendar(
                _replace=True,
                name='BASE',
                description='Base calendar',
                non_working_days=[date(2025, 1, 1), date(2025, 12, 25)],
            )
            base_cal.save()

            # Create adjustment that adds and removes days
            adj = CalendarAdjustment(
                _replace=True,
                name='TEST_ADJ',
                add_days=[date(2025, 1, 15)],
                remove_days=[date(2025, 1, 1)],
            )
            adj.save()

            # Create adjusted calendar
            adjusted_cal = Calendar(
                _replace=True,
                name='BASE',
                adjusted_for='TEST_ADJ',
            )
            adjusted_cal.save()

            # Adjusted calendar should have base days + add_days - remove_days
            nwds = set(adjusted_cal.non_working_days)
            assert date(2025, 1, 1) not in nwds  # removed
            assert date(2025, 1, 15) in nwds  # added
            assert date(2025, 12, 25) in nwds  # kept from base


class TestCalendar:
    """Unit tests for Calendar class."""

    def setup_method(self):
        """Set up test fixtures."""
        # Create test calendar with some holidays
        self.test_holidays = [
            date(2023, 1, 1),  # New Year's Day
            date(2023, 12, 25),  # Christmas
            date(2023, 7, 4),  # Independence Day
        ]

    def test_add_days(self):
        """Test add_days class method."""
        days = {date(2023, 1, 1), date(2023, 1, 2)}

        # Add new days
        result = Calendar.add_days(days, date(2023, 1, 3), date(2023, 1, 4))
        assert result is True
        assert days == {date(2023, 1, 1), date(2023, 1, 2), date(2023, 1, 3), date(2023, 1, 4)}

        # Add existing day (no change)
        result = Calendar.add_days(days, date(2023, 1, 1))
        assert result is False

        # Add no days
        result = Calendar.add_days(days)
        assert result is False

    def test_add_days_invalid_type(self):
        """Test add_days with invalid types raises TypeError."""
        days = set()
        with pytest.raises(TypeError, match='Every day to add must be a date'):
            Calendar.add_days(days, '2023-01-01')

    def test_remove_days(self):
        """Test remove_days class method."""
        days = {date(2023, 1, 1), date(2023, 1, 2), date(2023, 1, 3)}

        # Remove existing days
        result = Calendar.remove_days(days, date(2023, 1, 1), date(2023, 1, 2))
        assert result is True
        assert days == {date(2023, 1, 3)}

        # Remove non-existing day (no change)
        result = Calendar.remove_days(days, date(2023, 1, 4))
        assert result is False

        # Remove no days
        result = Calendar.remove_days(days)
        assert result is False

    def test_remove_days_invalid_type(self):
        """Test remove_days with invalid types raises TypeError."""
        days = {date(2023, 1, 1)}
        with pytest.raises(TypeError, match='Every day to remove must be a date'):
            Calendar.remove_days(days, '2023-01-01')

    def test_is_bizday(self):
        """Test is_bizday method using proper Traitable calendar."""
        with CACHE_ONLY():
            # Create a proper Calendar instance with non-working days
            cal = Calendar(
                name='TEST_CAL_HOLIDAYS', description='Calendar with holidays', non_working_days=[date(2023, 1, 1), date(2023, 12, 25)], _replace=True
            )

            # Test business days
            assert cal.is_bizday(date(2023, 1, 2)) is True
            assert cal.is_bizday(date(2023, 2, 1)) is True

            # Test non-business days
            assert cal.is_bizday(date(2023, 1, 1)) is False
            assert cal.is_bizday(date(2023, 12, 25)) is False

    def test_next_bizday(self):
        """Test next_bizday method using proper Traitable calendar."""
        with CACHE_ONLY():
            # Create a calendar with consecutive holidays
            cal = Calendar(
                name='TEST_CAL_CONSECUTIVE',
                description='Calendar with consecutive holidays',
                non_working_days=[
                    date(2023, 1, 1),  # Sunday
                    date(2023, 1, 2),  # Monday
                    date(2023, 1, 3),  # Tuesday
                ],
                _replace=True,
            )

            # Test from a business day
            result = cal.next_bizday(date(2023, 1, 4))  # Wednesday
            assert result == date(2023, 1, 5)  # Thursday

            # Test from a non-business day
            result = cal.next_bizday(date(2023, 1, 1))  # Sunday (holiday)
            assert result == date(2023, 1, 4)  # Wednesday

            # Test from a non-business day in middle of holidays
            result = cal.next_bizday(date(2023, 1, 2))  # Monday (holiday)
            assert result == date(2023, 1, 4)  # Wednesday

    def test_prev_bizday(self):
        """Test prev_bizday method using proper Traitable calendar."""
        with CACHE_ONLY():
            # Create a calendar with consecutive holidays
            cal = Calendar(
                name='TEST_CAL_PREV',
                description='Calendar for prev_bizday test',
                non_working_days=[
                    date(2023, 1, 1),  # Sunday
                    date(2023, 1, 2),  # Monday
                    date(2023, 1, 3),  # Tuesday
                ],
                _replace=True,
            )

            # Test from a business day
            result = cal.prev_bizday(date(2023, 1, 4))  # Wednesday
            assert result == date(2022, 12, 31)  # Saturday (goes back one day from Wednesday)

            # Test from a non-business day
            result = cal.prev_bizday(date(2023, 1, 1))  # Sunday (holiday)
            assert result == date(2022, 12, 31)  # Saturday

    def test_advance_bizdays_forward(self):
        """Test advance_bizdays with positive count using proper Traitable calendar."""
        with CACHE_ONLY():
            # Create a calendar with weekends and holidays
            cal = Calendar(
                name='TEST_CAL_ADVANCE',
                description='Calendar for advance_bizdays test',
                non_working_days=[
                    date(2023, 1, 1),  # New Year
                    date(2023, 1, 7),  # Saturday
                    date(2023, 1, 8),  # Sunday
                    date(2023, 1, 14),  # Saturday
                    date(2023, 1, 15),  # Sunday
                ],
                _replace=True,
            )

            # Start from Friday Jan 6, 2023
            start_date = date(2023, 1, 6)  # Friday

            # Advance 1 business day (skip weekend)
            result = cal.advance_bizdays(start_date, 1)
            assert result == date(2023, 1, 9)  # Monday

            # Advance 3 business days
            result = cal.advance_bizdays(start_date, 3)
            assert result == date(2023, 1, 11)  # Wednesday

            # Advance 0 business days (no change)
            result = cal.advance_bizdays(start_date, 0)
            assert result == start_date

    def test_advance_bizdays_backward(self):
        """Test advance_bizdays with negative count using proper Traitable calendar."""
        with CACHE_ONLY():
            # Create a calendar with weekends and holidays (same as forward test)
            cal = Calendar(
                name='TEST_CAL_BACKWARD',
                description='Calendar for backward advance_bizdays test',
                non_working_days=[
                    date(2023, 1, 1),  # New Year
                    date(2023, 1, 7),  # Saturday
                    date(2023, 1, 8),  # Sunday
                    date(2023, 1, 14),  # Saturday
                    date(2023, 1, 15),  # Sunday
                ],
                _replace=True,
            )

            # Start from Wednesday Jan 11, 2023
            start_date = date(2023, 1, 11)  # Wednesday

            # Go back 1 business day (skip weekend)
            result = cal.advance_bizdays(start_date, -1)
            assert result == date(2023, 1, 10)  # Tuesday

            # Go back 3 business days
            result = cal.advance_bizdays(start_date, -3)
            assert result == date(2023, 1, 6)  # Friday

    def test_and(self):
        """Test AND class method."""
        # This would require actual calendar instances
        # Test the None case
        result = Calendar.AND()
        assert result is None

    def test_or(self):
        """Test OR class method."""
        # This would require actual calendar instances
        # Test the None case
        result = Calendar.OR()
        assert result is None

    def test_union(self):
        """Test union class method."""
        # Test with no calendars
        with pytest.raises(AssertionError, match='At least one calendar is required for union'):
            Calendar.union()

    def test_calendar_union_and_intersection_real_calendars(self, ts_instance):
        """Test Calendar.union and Calendar.intersection on real stored calendars."""
        with ts_instance:
            us_holidays = {
                date(2025, 1, 1),
                date(2025, 1, 20),
                date(2025, 2, 17),
                date(2025, 5, 26),
                date(2025, 7, 4),
                date(2025, 9, 1),
                date(2025, 10, 13),
                date(2025, 11, 11),
                date(2025, 11, 27),
                date(2025, 12, 25),
            }
            uk_holidays = {
                date(2025, 1, 1),
                date(2025, 4, 18),
                date(2025, 5, 5),
                date(2025, 5, 26),
                date(2025, 8, 25),
                date(2025, 12, 25),
                date(2025, 12, 26),
            }

            us_c = Calendar(
                _replace=True,
                name='US',
                description='US calendar',
                non_working_days=sorted(us_holidays),
            )
            uk_c = Calendar(
                _replace=True,
                name='UK',
                description='UK calendar',
                non_working_days=sorted(uk_holidays),
            )

            assert us_c.save()
            assert uk_c.save()

            union_cal = Calendar.union('US', 'UK')
            assert isinstance(union_cal, Calendar)
            assert set(union_cal.non_working_days) == us_holidays | uk_holidays

            intersection_cal = Calendar.intersection('US', 'UK')
            assert isinstance(intersection_cal, Calendar)
            assert set(intersection_cal.non_working_days) == us_holidays & uk_holidays

    def test_intersection_alias(self):
        """Test that intersection is an alias for AND."""
        assert Calendar.intersection == Calendar.AND

    def test_add_non_working_days(self):
        """Test add_non_working_days instance method."""
        # Test the class method that add_non_working_days would use
        days = {date(2023, 1, 1)}
        Calendar.add_days(days, date(2023, 12, 25), date(2023, 7, 4))
        assert days == {date(2023, 1, 1), date(2023, 12, 25), date(2023, 7, 4)}

    def test_remove_non_working_days(self):
        """Test remove_non_working_days instance method."""
        # Test the class method that remove_non_working_days would use
        days = {date(2023, 1, 1), date(2023, 12, 25), date(2023, 7, 4)}
        Calendar.remove_days(days, date(2023, 12, 25))
        assert days == {date(2023, 1, 1), date(2023, 7, 4)}

    def test_non_working_days_get(self):
        """Test non_working_days_get method."""
        # This method requires database access, so we'll skip detailed testing
        # The method parses calendar names and gets holidays from database
        pass

    def test_non_working_days_trait_get(self):
        """Test _non_working_days_get trait method."""
        # Test the logic that _non_working_days_get would use
        non_working_days = [date(2023, 1, 1), date(2023, 12, 25)]
        result = set(non_working_days)
        assert result == {date(2023, 1, 1), date(2023, 12, 25)}

    def test_calendar_instantiation(self):
        """Test proper Calendar instantiation following Traitable patterns."""
        # Calendar is storable, so it needs proper trait values like other storable classes
        with CACHE_ONLY():
            # Calendar requires name (which has T.ID flag)
            cal = Calendar(name='TEST_CAL', description='Test Calendar', _replace=True)
            assert cal.name == 'TEST_CAL'
            assert cal.description == 'Test Calendar'

    def test_calendar_is_storable(self):
        """Test that Calendar is storable like other Traitable subclasses."""
        assert Calendar.is_storable()
        assert Calendar.trait('name')  # Has ID trait
        assert Calendar.trait('_collection_name')  # Has collection name trait



from __future__ import annotations

import functools
from functools import partial

import math
import re
from datetime import date

from dateutil.relativedelta import relativedelta as delta

from core_10x.named_constant import NamedConstant, NamedConstantTable
from core_10x.nucleus import Nucleus
from xx_common.xxcalendar import Calendar

#=====
#   Biz Day Roll Rules
#   1. Preceding (rolls backwards): go to the prev biz day, if not already
#   2. Following (rolls forward):   go to the next biz day, if not already
#   3. Modified following:          go to the next biz day unless the next biz day is in the next month, otherwise go to 1.
#   4. Modified preceding:          go to the prev biz day unless the prev biz day is in the prev month, otherwise go to 2.
#=====
def bizday_roll_preceding(d: date, cal: Calendar) -> date:
    if cal.is_bizday(d):
        return d

    return cal.prev_bizday(d)


def bizday_roll_following(d: date, cal: Calendar) -> date:
    if cal.is_bizday(d):
        return d

    return cal.next_bizday(d)


def bizday_roll_mod_following(d: date, cal: Calendar) -> date:
    if cal.is_bizday(d):
        return d

    res = cal.next_bizday(d)
    return res if res.month == d.month else cal.prev_bizday(d)


def bizday_roll_mod_preceding(d: date, cal: Calendar) -> date:
    if cal.is_bizday(d):
        return d

    res = cal.prev_bizday(d)
    return res if res.month == d.month else cal.next_bizday(d)

class BIZDAY_ROLL_RULE(NamedConstant):
    PRECEDING       = bizday_roll_preceding
    FOLLOWING       = bizday_roll_following
    MOD_FOLLOWING   = bizday_roll_mod_following
    MOD_PRECEDING   = bizday_roll_mod_preceding
    NO_ROLL         = lambda d, cal: d

class TENOR_FREQUENCY(NamedConstant):
    BIZDAY      = ()
    CALDAY      = ()
    WEEK        = ()
    MONTH       = ()
    QUARTER     = ()
    HALF_YEAR   = ()
    YEAR        = ()
    SOM         = ()
    EOM         = ()
    #IMM_QUARTER = ()   # TODO - use later

class TENOR_PARAMS(NamedConstant):
    CHAR            = ()
    RDATE_FUNC      = ()
    CONVERSIONS     = ()
    MIN_FREQUENCY   = ()

FREQUENCY_TABLE = NamedConstantTable(TENOR_FREQUENCY, TENOR_PARAMS,
    #              CHAR         RDATE_FUNC                                  CONVERSIONS                                         MIN_FREQUENCY
    BIZDAY      = ('B',         None,                                       None,                                               None),
    CALDAY      = ('C',         lambda d, n: d + delta(days    = n),        dict(C = 1,     W = 1/7),                           TENOR_FREQUENCY.CALDAY),
    WEEK        = ('W',         lambda d, n: d + delta(weeks   = n),        dict(C = 7,     W = 1),                             TENOR_FREQUENCY.CALDAY),
    MONTH       = ('M',         lambda d, n: d + delta(months  = n),        dict(M = 1,     Q = 1/3,    S = 1/6,    Y = 1/12),  TENOR_FREQUENCY.MONTH),
    QUARTER     = ('Q',         lambda d, n: d + delta(months  = n * 3),    dict(M = 3,     Q = 1,      S = 0.5,    Y = 0.25),  TENOR_FREQUENCY.MONTH),
    HALF_YEAR   = ('S',         lambda d, n: d + delta(months  = n * 6),    dict(M = 6,     Q = 2,      S = 1,      Y = 0.5),   TENOR_FREQUENCY.MONTH),
    YEAR        = ('Y',         lambda d, n: d + delta(years   = n),        dict(M = 12,    Q = 4,      S = 2,      Y = 1),     TENOR_FREQUENCY.MONTH),

    SOM         = ('SOM',       lambda d, n: date(d.year, d.month, 1) + delta(months = n),                          None,       None),
    EOM         = ('EOM',       lambda d, n: date(d.year, d.month, 1) + delta(months = n+1) + delta(days = -1),     None,       None),

    #IMM_QUARTER = ( 'IMM',      lambda d, n: IMMQuarter.which( d ).ith( n ).last(), None,                                               None ),
)

class PROPAGATE_DATES(NamedConstant):
    BACKWARD = ()  ## BACKWARD period propagation is more standard for G10 interest rate swaps
    FORWARD = ()

class RDate(Nucleus):
    """
    rd = RDate('1Y')
    dt = date(2025, 1, 1)
    rd.apply(dt, Calendar.existing_instance_by_id('FB'), BIZDAY_ROLL_RULE.FOLLOWING)
    -> date(2026, 1, 1)
    """

    s_tenor_frequency_class = TENOR_FREQUENCY
    s_frequency_table       = FREQUENCY_TABLE
    def __init_subclass__(cls, tenor_frequency_class: type[TENOR_FREQUENCY] = None, frequency_table: NamedConstantTable = None):
        assert tenor_frequency_class and issubclass(tenor_frequency_class, TENOR_FREQUENCY), 'tenor_frequency_class must be a subclass of TENOR_FREQUENCY'
        assert frequency_table and isinstance(frequency_table, NamedConstantTable) and frequency_table.named_constant_class is tenor_frequency_class, f"""
        frequency_table's named_constant_class must be {tenor_frequency_class}
        """
        cls.s_tenor_frequency_class = tenor_frequency_class
        cls.s_frequency_table = frequency_table

    # s_spot_tenors = { 'ON', 'TN', 'SN' }

    # -- RDate('3M') or RDate(TENOR_FREQUENCY.MONTH, 3)
    # def __init__(self, symbol_or_frequency = None, count: int = None):
    def __init__(self, symbol: str = None, freq: TENOR_FREQUENCY = None, count: int = None):
        cls = self.__class__
        if symbol is not None:
            match = re.match(r'(-?\d+)([a-zA-Z]+)', symbol)
            count = int(match.group(1))
            letter = match.group(2)

            freq = cls.s_frequency_table.primary_key(TENOR_PARAMS.CHAR, letter.upper())
            if freq is None:
                raise AttributeError(f"Invalid tenor symbol '{symbol}'")

        else:
            assert freq is not None, 'freq must be a valid TENOR_FREQUENCY'
            if count is None:
                count = 1

        self.freq = freq
        self.count = count

    def symbol(self) -> str:
        return f'{self.count}{self.__class__.s_frequency_table[self.freq][TENOR_PARAMS.CHAR]}'

    #---- Nucleus methods
    def to_str(self) -> str:
        return self.symbol()

    def serialize(self, embed: bool):
        return self.symbol()

    @classmethod
    def deserialize(cls, serialized_data) -> Nucleus:
        return cls(symbol = serialized_data)

    @classmethod
    def from_str(cls, s: str) -> Nucleus:
        return cls(symbol = s)

    @classmethod
    def from_any_xstr(cls, value) -> Nucleus:
        if isinstance(value, cls):
            return cls(freq = value.freq, count = value.count)

        if isinstance(value, tuple):
            if len(value) != 2:
                raise AssertionError('tenor frequency and count are expected')
            return cls(freq = value[0], count = value[1])

        raise AssertionError('unexpected type')

    @classmethod
    def same_values(cls, value1, value2) -> bool:
        return value1.freq == value2.freq and value1.count == value2.count

    def apply(self, d: date, cal: Calendar, roll_rule: BIZDAY_ROLL_RULE) -> date:
        cls = self.__class__
        if self.freq is cls.s_tenor_frequency_class.BIZDAY:
            not_rolled = cal.advance_bizdays(d, self.count)     ## roll to bizday for '0B'
        else:
            fn = cls.s_frequency_table[self.freq][TENOR_PARAMS.RDATE_FUNC]
            not_rolled = fn(d, self.count)

        return roll_rule(not_rolled, cal)

    @classmethod
    def apply_rule(cls, d: date, cal: Calendar, roll_rule: BIZDAY_ROLL_RULE, comma_separated_rdates_str: str) -> date:
        rdates = comma_separated_rdates_str.split(',')
        for rdate_symbol in rdates:
            rdate = RDate(symbol = rdate_symbol)
            d = rdate.apply(d, cal, roll_rule)
        return d

    def conversion_freq_multiplier(self, other_freq: TENOR_FREQUENCY) -> float:
        my_freq = self.freq
        if my_freq is other_freq:
            return 1.0

        freq_table = self.__class__.s_frequency_table
        conversion_map = freq_table[my_freq][TENOR_PARAMS.CONVERSIONS]
        if not conversion_map:
            raise ValueError(f'{my_freq} cannot be converted into any other frequency')

        multiplier = conversion_map.get(freq_table[other_freq][TENOR_PARAMS.CHAR])
        if not multiplier:
            raise ValueError(f'cannot get a conversion multiplier from {my_freq} to {other_freq}')

        return multiplier

    def equate_freq(self, other: RDate) -> tuple:
        if self.freq is other.freq:
            return (self, other)

        mult = self.conversion_freq_multiplier(other.freq)  # -- either int or 1/int
        mult_i = int(mult)
        if mult_i >= 1:
            return (RDate(freq = other.freq, count = self.count * mult_i), other)

        return (self, RDate(freq = self.freq, count = other.count * int(1 / mult)))

    def _fract_count_to_RDate(self, count: float, freq: TENOR_FREQUENCY, freq_to_try: TENOR_FREQUENCY) -> RDate:
        if math.isclose(count, round(count)):
            return RDate(freq = freq, count = round(count))

        conv_to_min_freq = self.conversion_freq_multiplier(freq_to_try)
        return RDate(freq = freq_to_try, count = round(conv_to_min_freq * count))

    def multadd(self, mult: float, add, freq_to_try_for_fract_count = None) -> RDate:
        if isinstance(add, (int, float)):
            if add:
                raise ValueError(f'cannot add a non-zero number {add} to RDate {self}: the result is ambiguous')

            return self._fract_count_to_RDate(mult * self.count, self.freq, freq_to_try_for_fract_count)

        if isinstance(add, str):
            try:
                add = RDate(add)
            except Exception as e:
                raise ValueError(f'cannot convert adder {add} to RDate') from e

        if isinstance(add, RDate):
            rd1, rd2 = self.equate_freq(add)
            return self._fract_count_to_RDate(mult * rd1.count + rd2.count, rd1.freq, freq_to_try_for_fract_count)

        raise ValueError(f'cannot calc a linear combination of {mult} * {self} + {add}')

    def __mul__(self, other: float) -> RDate:
        return self.multadd(other, 0, freq_to_try_for_fract_count = self.__class__.s_frequency_table[self.freq][TENOR_PARAMS.MIN_FREQUENCY])

    def __rmul__(self, other: float) -> RDate:
        return self.multadd(other, 0, freq_to_try_for_fract_count=self.__class__.s_frequency_table[self.freq][TENOR_PARAMS.MIN_FREQUENCY])

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return self.multadd(1 / other, 0, freq_to_try_for_fract_count=self.__class__.s_frequency_table[self.freq][TENOR_PARAMS.MIN_FREQUENCY])

        rd1, rd2 = self.equate_freq(other)
        return rd1.count / rd2.count

    def __add__(self, other) -> RDate:
        return self.multadd(1.0, other, freq_to_try_for_fract_count=self.__class__.s_frequency_table[self.freq][TENOR_PARAMS.MIN_FREQUENCY])

    @classmethod
    def add_bizdays(cls, d: date, biz_days: int, cal: Calendar, roll_rule: BIZDAY_ROLL_RULE) -> date:
        not_rolled = cal.advance_bizdays(d, biz_days)
        return roll_rule(not_rolled, cal)

    @classmethod
    def roll_to_bizday(cls, d: date, cal: Calendar, roll_rule: BIZDAY_ROLL_RULE) -> date:
        return roll_rule(d, cal)

    class RELOP(NamedConstant):
        LT = date.__lt__
        GT = date.__gt__
        EQ = date.__eq__
        NE = date.__ne__
        LE = date.__le__
        GE = date.__ge__

    @classmethod
    def relop(cls, rel_op: RELOP, rd1: RDate, rd2: RDate, d: date, cal: Calendar, roll_rule: BIZDAY_ROLL_RULE) -> bool:
        d1 = rd1.apply(d, cal, roll_rule)
        d2 = rd2.apply(d, cal, roll_rule)
        return rel_op(d1, d2)

    @classmethod
    def from_tenors(cls, tenors_str: str, delim = ',') -> list:
        tenors = tenors_str.split(delim)
        return [RDate(tenor.strip()) for tenor in tenors]

    def dates_schedule(self, start: date, end: date, calendar: Calendar, roll_rule: BIZDAY_ROLL_RULE, date_propagation: PROPAGATE_DATES, allow_stub = True) -> list:
        rolled_start = roll_rule(start, calendar)
        if rolled_start < start:
            raise ValueError(f'rolled start date {rolled_start} is before non-working start date {start} according to calendar {calendar.name} and roll rule {roll_rule.name}')

        rolled_end = roll_rule(end, calendar)
        if rolled_end > end:
            raise ValueError(f'rolled end date {rolled_end} is after non-working end date {end} according to calendar {calendar.name} and roll rule {roll_rule.name}')

        if date_propagation == PROPAGATE_DATES.BACKWARD:
            begin = rolled_end
            finish = rolled_start
            dir = -1
            exceed = lambda x, y: x < y
            ex_or_eq = lambda x, y: x <= y
        else:
            begin = rolled_start
            finish = rolled_end
            dir = 1
            exceed = lambda x, y: x > y
            ex_or_eq = lambda x, y: x >= y

        step = self * dir

        prev_rolled_date = begin
        rolled_date = begin
        non_rolled_date = begin

        all_dates = []
        while ex_or_eq(finish, rolled_date):
            all_dates.append(rolled_date)
            non_rolled_date = step.apply(non_rolled_date, calendar, BIZDAY_ROLL_RULE.NO_ROLL)
            rolled_date = roll_rule(non_rolled_date, calendar)
            if not exceed(rolled_date, prev_rolled_date):
                raise ValueError(f'infinite loop: next date {rolled_date} vs previous date {prev_rolled_date} for date_propagation = {date_propagation.name}')
            if exceed(rolled_date, finish):
                break
            prev_rolled_date = rolled_date

        if not allow_stub:
            if prev_rolled_date != finish:
                raise ValueError(f'the date sequence has a stub period bound by {prev_rolled_date} and {finish}')
        else:
            if finish not in all_dates:
                all_dates.append(finish)

        all_dates.sort()
        return all_dates

    def period_dates(self, start: date, end: date, calendar: Calendar, roll_rule: BIZDAY_ROLL_RULE, date_propagation: PROPAGATE_DATES, allow_stub = True) -> tuple:
        all_dates = self.dates_schedule(start, end, calendar, roll_rule, date_propagation, allow_stub)
        start_dates = all_dates[:-1]
        end_dates = all_dates[1:]

        return start_dates, end_dates, all_dates

    ## if the tenor is not an integral multiple of the frequency (e.g., 30-month annual swap: tenor = 5S, freq = YEAR)
    ## then there would be a period "shorter than the frequency" (e.g., half-year in the 30-month annual swap)
    ## call such a period a STUB. it maybe in the front (1st period) for BACKWARD propagation or in the back (last period) for FORWARD
    def period_dates_for_tenor(self, start: date, tenor: RDate, calendar: Calendar, roll_rule: BIZDAY_ROLL_RULE, date_propagation: PROPAGATE_DATES, allow_stub = True) -> tuple:
        start = roll_rule(start, calendar)
        end = tenor.apply(start, calendar, roll_rule)
        return self.period_dates(start, end, calendar, roll_rule, date_propagation, allow_stub)

    @classmethod
    def bizdays_for_period(cls, start_date: date, end_date: date, calendar: Calendar) -> list:
        if end_date < start_date:
            start_date, end_date = end_date, start_date

        bizdays = []
        d = RDate.roll_to_bizday(start_date, calendar, BIZDAY_ROLL_RULE.FOLLOWING)
        while d <= end_date:
            bizdays.append(d)
            d = RDate.add_bizdays(d, 1, calendar, BIZDAY_ROLL_RULE.NO_ROLL)

        return bizdays

    @classmethod
    def apply_bound(cls, rdate: RDate = None, cal: Calendar = None, roll_rule: BIZDAY_ROLL_RULE = None):
        kwargs = {}
        if rdate is not None:
            kwargs['rdate'] = rdate
        if cal is not None:
            kwargs['cal'] = cal
        if roll_rule is not None:
            kwargs['roll_rule'] = roll_rule
        return functools.partial(cls.apply, **kwargs)

    @classmethod
    def prior_to(cls, ordered_rdates: list, start_date: date, end_date: date, cal: Calendar, roll_rule: BIZDAY_ROLL_RULE, overstep = True) -> list:
        if start_date > end_date or not ordered_rdates:
            return []

        rd: RDate
        subset = [ (rd, d, i) for i, rd in enumerate(ordered_rdates) if (d := rd.apply(start_date, cal, roll_rule)) <= end_date ]

        if len(subset) == len(ordered_rdates):  ## last d < end_date
            return ordered_rdates

        if not subset:
            return ordered_rdates[0] if overstep else []

        if overstep and subset[-1][1] < end_date:
            i = subset[-1][2] + 1   #-- step over
            subset.append((ordered_rdates[i], None, -1))  ## (rd, _, _)

        return [rd for (rd, _, _) in subset]



from core_10x.named_constant import NamedConstant

# fmt: off
BACKBONE_STORE_CLASS_NAME   = 'infra_10x.mongo_db.Mongo'
#BACKBONE_USER_NAME         = 'core_10x.backbone.backbone_admin.BACKBONE_USER'

AUTO_PASSWORD_LENGTH        = 32

class DB_NAME:
    ANY         = ''
    NONE        = '_none'
    ADMIN       = 'admin'
    BACKBONE    = '_backbone_'
    VAULT       = '_vault_'
    SHADOW      = '_shadow_'

class STANDARD_ROLE:
    READ            = 'read'
    READ_ANY        = 'readAnyDatabase'
    READ_WRITE      = 'readWrite'
    READ_WRITE_ANY  = 'readWriteAnyDatabase'
    DB_ADMIN        = 'dbAdmin'
    DB_ADMIN_ANY    = 'dbAdminAnyDatabase'
    OWNER           = 'dbOwner'
    USER_ADMIN      = 'userAdmin'
    USER_ADMIN_ANY  = 'userAdminAnyDatabase'
    CLUSTER_ADMIN   = 'clusterAdmin'
    CLUSTER_MAN     = 'clusterManager'
    BACKUP          = 'backup'
    RESTORE         = 'restore'
    ROOT            = 'root'

class ACTION:
    FIND            = 'find'
    INSERT          = 'insert'
    UPDATE          = 'update'
    REMOVE          = 'remove'
    INDEX_LIST      = 'listIndexes'
    INDEX_CREATE    = 'createIndex'
    COLL_LIST       = 'listCollections'
    COLL_STATS      = 'collStats'
    COLL_RENAME     = 'renameCollectionSameDB'
    COLL_CREATE     = 'createCollection'
    COLL_DROP       = 'dropCollection'
    DB_LIST         = 'listDatabases'
    DB_STATS        = 'dbStats'
    CHANGE_STREAME  = 'changeStream'

    CHANGE_OWN_PASSWORD = 'changeOwnPassword'

class PERMISSION:
    NONE        = []
    NEW_PWD     = [ ACTION.CHANGE_OWN_PASSWORD ]
    READ_ONLY   = [ ACTION.FIND ]
    READ_EX     = [ ACTION.FIND, ACTION.COLL_LIST ]
    WRITE       = [ ACTION.INSERT, ACTION.UPDATE, ACTION.REMOVE ]
    INDEX       = [ ACTION.INDEX_LIST, ACTION.INDEX_CREATE ]
    COLL        = [ ACTION.COLL_LIST, ACTION.COLL_RENAME, ACTION.COLL_STATS, ACTION.COLL_CREATE, ACTION.COLL_DROP ]

    UPDATE      = [ *READ_ONLY, ACTION.UPDATE ]
    READ_WRITE  = [ *READ_ONLY, *WRITE, *INDEX, *COLL ]

USER_ADMIN_SUFFIX               = 'admin'
FUNCTIONAL_ACCOUNT_PREFIX       = 'xx'
FUNCTIONAL_ACCOUNT_DEPARTMENT   = 'XX Functional Accounts'

class USER_ROLE(NamedConstant):
    NONE        = []
    VISITOR     = PERMISSION.READ_EX
    WORKER      = PERMISSION.READ_WRITE

class COLL_PLACEHOLDER:
    TAG         = '__coll_placeholder'
    USERNAME    = 'username'
# fmt: on


class CUSTOM_ROLE(NamedConstant):
    """
    role_name: e.g., 'VISITOR'
    permissions_per_db_name: a dict: each key is either a db_name or db_name/collection_name, value is PERMISSION, e.g.,
    { 'catalog': PERMISSION.READ_ONLY, 'logs/errors': PERMISSION.READ_WRITE }
    """

    @classmethod
    def coll_placeholder(cls, value) -> tuple:  # -- (coll_placeholder, stripped_value)
        cvalue = dict(value)
        return (cvalue.pop(COLL_PLACEHOLDER.TAG, None), value)

    @classmethod
    def user_role_name(cls, username: str) -> str:
        return f'{username.upper()}_USR'


# fmt: off
SECURITY_KEYS_MISSING = 'Sec keys are missing. If you are using a new computer, please run "xx user new machine" from your shell'
SECURITY_KEYS_INCOMPATIBLE = 'Sec keys are incompatible with an encrypted password'

SECRETS_CLIENT_ARGS     = dict(service_name = 'secretmanager', region_name = 'me-south-1')
SECRETS_PATH_PREFIX     = '/external-secrets/xx'
# fmt: on



import secrets

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.serialization import load_pem_private_key, load_pem_public_key

# fmt: off
PUBLIC_EXP  = 65537
KEY_SIZE    = 2048
PWD_SIZE    = 24
ENCODING    = 'utf-8'
# fmt: on


class SecKeys:
    @classmethod
    def generate_password(cls, length=PWD_SIZE) -> str:
        return secrets.token_urlsafe(length)

    @classmethod
    def generate_keys(cls, pwd=None) -> tuple:
        private_key = rsa.generate_private_key(public_exponent=PUBLIC_EXP, key_size=KEY_SIZE, backend=default_backend())
        public_key = private_key.public_key()

        if pwd:
            format = serialization.PrivateFormat.PKCS8
            algo = serialization.BestAvailableEncryption(bytes(pwd, encoding=ENCODING))
        else:
            format = serialization.PrivateFormat.TraditionalOpenSSL
            algo = serialization.NoEncryption()

        private_key_pem = private_key.private_bytes(encoding=serialization.Encoding.PEM, format=format, encryption_algorithm=algo)
        public_key_pem = public_key.public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo)

        return (private_key_pem, public_key_pem)

    @classmethod
    def encrypt(cls, message, public_key_pem: bytes) -> bytes:
        if type(message) is str:
            message = bytes(message, encoding=ENCODING)

        public_key = load_pem_public_key(public_key_pem)
        # fmt: off
        return public_key.encrypt(
            message,
            padding.OAEP(
                mgf         = padding.MGF1(algorithm = hashes.SHA256()),
                algorithm   = hashes.SHA256(),
                label       = None
            )
        )
        # fmt: on

    @classmethod
    def decrypt(cls, encrypted_message: bytes, private_key_pem: bytes, to_str=True):
        private_key = load_pem_private_key(private_key_pem, password=None)
        # fmt: off
        res = private_key.decrypt(
            encrypted_message,
            padding.OAEP(
                mgf         = padding.MGF1(algorithm = hashes.SHA256()),
                algorithm   = hashes.SHA256(),
                label       = None
            )
        )
        # fmt: on

        if to_str:
            res = res.decode(encoding=ENCODING)

        return res

    @classmethod
    def encrypt_private_key(cls, private_key_pem: bytes, password) -> bytes:
        if type(password) is str:
            password = bytes(password, encoding=ENCODING)

        private_key = load_pem_private_key(private_key_pem, password=None)
        # fmt: off
        return private_key.private_bytes(
            encoding                = serialization.Encoding.PEM,
            format                  = serialization.PrivateFormat.PKCS8,
            encryption_algorithm    = serialization.BestAvailableEncryption(password),
        )
        # fmt: on

    @classmethod
    def decrypt_private_key(cls, private_key_with_password, password) -> bytes:
        if type(password) is str:
            password = bytes(password, encoding=ENCODING)

        pk = load_pem_private_key(private_key_with_password, password=password)
        # fmt: off
        return pk.private_bytes(
            encoding                = serialization.Encoding.PEM,
            format                  = serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm    = serialization.NoEncryption(),
        )
        # fmt: on

    def __init__(self, private_key_with_password: bytes, public_key_pem: bytes, password):
        if type(password) is str:
            password = bytes(password, encoding=ENCODING)

        self.private_key = load_pem_private_key(private_key_with_password, password=password)
        self.public_key = load_pem_public_key(public_key_pem)

    def encrypt_text(self, text: str) -> bytes:
        message = bytes(text, encoding=ENCODING)
        # fmt: off
        return self.public_key.encrypt(
            message,
            padding.OAEP(
                mgf         = padding.MGF1(algorithm = hashes.SHA256()),
                algorithm   = hashes.SHA256(),
                label       = None
            )
        )

    def decrypt_text(self, encrypted_message: bytes) -> str:
        # fmt: off
        res = self.private_key.decrypt(
            encrypted_message,
            padding.OAEP(
                mgf         = padding.MGF1(algorithm = hashes.SHA256()),
                algorithm   = hashes.SHA256(),
                label       = None
            )
        )
        # fmt: on

        return res.decode(encoding=ENCODING)



"""
Tests for ui_10x.rio.internals module.

This module tests the App10x, FastapiServer, and Session classes that provide
custom Rio app functionality for running in webview windows.
"""

import pathlib
import sys
from unittest.mock import Mock, patch

import pytest
from ui_10x.rio.internals.app import App10x, FastapiServer, Session, app_server

import rio


class TestApp10x:
    """Test cases for the App10x class."""

    def test_app10x_initialization(self):
        """Test App10x initialization with default values."""
        mock_app = Mock()
        app10x = App10x(app=mock_app)

        assert app10x.app is mock_app
        assert app10x.webview is None

    def test_app10x_initialization_with_webview(self):
        """Test App10x initialization with webview provided."""
        mock_app = Mock()
        mock_webview = Mock()
        app10x = App10x(app=mock_app, webview=mock_webview)

        assert app10x.app is mock_app
        assert app10x.webview is mock_webview

    def test_update_window_size_no_dimensions(self, monkeypatch):
        """Test _update_window_size when no dimensions are provided."""
        # Mock the webview module that gets imported inside the method
        mock_webview = Mock()
        mock_webview.windows = [Mock()]
        monkeypatch.setitem(sys.modules, 'webview', mock_webview)

        App10x._update_window_size(None, None)
        # Should return early without calling webview methods
        # Since webview is imported inside the method, we can't easily test the early return
        # without more complex mocking, so we'll just verify the method doesn't crash
        pass

    def test_update_window_size_with_dimensions(self, monkeypatch):
        """Test _update_window_size when dimensions are provided."""
        mock_webview = Mock()
        mock_window = Mock()
        mock_window.width = 800
        mock_window.height = 600
        mock_window.evaluate_js.return_value = 16.0  # pixels_per_rem
        mock_webview.windows = [mock_window]

        monkeypatch.setitem(sys.modules, 'webview', mock_webview)

        App10x._update_window_size(50.0, 40.0)

        # Verify evaluate_js was called to get pixels_per_rem
        mock_window.evaluate_js.assert_called_once()
        # Verify resize was called with calculated pixel dimensions
        mock_window.resize.assert_called_once_with(800, 640)  # 50*16, 40*16

    def test_update_window_size_partial_dimensions(self, monkeypatch):
        """Test _update_window_size with only width or height provided."""
        mock_webview = Mock()
        mock_window = Mock()
        mock_window.width = 800
        mock_window.height = 600
        mock_window.evaluate_js.return_value = 16.0
        mock_webview.windows = [mock_window]

        monkeypatch.setitem(sys.modules, 'webview', mock_webview)

        # Test with only width
        App10x._update_window_size(50.0, None)
        mock_window.resize.assert_called_with(800, 600)  # width calculated, height unchanged

        # Reset for next test
        mock_window.reset_mock()
        mock_window.evaluate_js.return_value = 16.0

        # Test with only height
        App10x._update_window_size(None, 40.0)
        mock_window.resize.assert_called_with(800, 640)  # height calculated, width unchanged

    @patch('ui_10x.rio.internals.app.WebViewProcess')
    @patch('ui_10x.rio.internals.app.utils.ensure_valid_port')
    @patch('asyncio.run')
    def test_run_in_window_basic(self, mock_asyncio_run, mock_ensure_port, mock_webview_process):
        """Test basic _run_in_window functionality."""
        # Setup mocks
        mock_ensure_port.return_value = 8080
        mock_asyncio_run.return_value = '/path/to/icon.png'
        mock_webview = Mock()
        mock_webview.is_alive.return_value = True
        mock_webview_process.return_value = mock_webview

        mock_app = Mock()
        mock_app.name = 'Test App'
        app10x = App10x(app=mock_app)

        # Mock the _run_as_web_server method
        mock_app._run_as_web_server = Mock()

        app10x._run_in_window()

        # Verify WebViewProcess was created with correct parameters
        mock_webview_process.assert_called_once()
        call_kwargs = mock_webview_process.call_args[1]
        assert call_kwargs['url'] == 'http://localhost:8080'
        assert call_kwargs['title'] == 'Test App'
        assert call_kwargs['maximized'] is False
        assert call_kwargs['fullscreen'] is False

        # Verify _run_as_web_server was called
        mock_app._run_as_web_server.assert_called_once()
        call_kwargs = mock_app._run_as_web_server.call_args[1]
        assert call_kwargs['host'] == 'localhost'
        assert call_kwargs['port'] == 8080
        assert call_kwargs['running_in_window'] is True
        assert 'internal_on_server_created' in call_kwargs

    @patch('ui_10x.rio.internals.app.WebViewProcess')
    @patch('ui_10x.rio.internals.app.utils.ensure_valid_port')
    @patch('asyncio.run')
    def test_run_in_window_with_options(self, mock_asyncio_run, mock_ensure_port, mock_webview_process):
        """Test _run_in_window with various options."""
        mock_ensure_port.return_value = 8080
        mock_asyncio_run.return_value = '/path/to/icon.png'
        mock_webview = Mock()
        mock_webview.is_alive.return_value = True
        mock_webview_process.return_value = mock_webview

        mock_app = Mock()
        mock_app.name = 'Test App'
        app10x = App10x(app=mock_app)
        mock_app._run_as_web_server = Mock()

        # Test with custom options
        app10x._run_in_window(quiet=False, maximized=True, fullscreen=True, width=100.0, height=80.0, debug_mode=True)

        # Verify WebViewProcess was created with custom options
        call_kwargs = mock_webview_process.call_args[1]
        assert call_kwargs['maximized'] is True
        assert call_kwargs['fullscreen'] is True

        # Verify _run_as_web_server was called with custom options
        call_kwargs = mock_app._run_as_web_server.call_args[1]
        assert call_kwargs['quiet'] is False
        assert call_kwargs['debug_mode'] is True

    @patch('ui_10x.rio.internals.app.WebViewProcess')
    @patch('ui_10x.rio.internals.app.utils.ensure_valid_port')
    @patch('asyncio.run')
    def test_run_in_window_exception_handling(self, mock_asyncio_run, mock_ensure_port, mock_webview_process):
        """Test _run_in_window exception handling and cleanup."""
        mock_ensure_port.return_value = 8080
        mock_asyncio_run.return_value = '/path/to/icon.png'
        mock_webview = Mock()
        mock_webview.is_alive.return_value = True
        mock_webview_process.return_value = mock_webview

        mock_app = Mock()
        mock_app.name = 'Test App'
        mock_app._run_as_web_server.side_effect = Exception('Test error')
        app10x = App10x(app=mock_app)

        # Should not raise exception
        app10x._run_in_window()

        # Verify cleanup was performed
        mock_webview.close.assert_called_once()
        mock_webview.join.assert_called_once()

    @patch('ui_10x.rio.internals.app.WebViewProcess')
    @patch('ui_10x.rio.internals.app.utils.ensure_valid_port')
    @patch('asyncio.run')
    def test_run_in_window_on_server_created_callback(self, mock_asyncio_run, mock_ensure_port, mock_webview_process):
        """Test _on_server_created callback functionality."""
        # Setup mocks
        mock_ensure_port.return_value = 8080
        mock_asyncio_run.return_value = '/path/to/icon.png'
        mock_webview = Mock()
        mock_webview.is_alive.return_value = True
        mock_webview_process.return_value = mock_webview

        mock_app = Mock()
        mock_app.name = 'Test App'
        app10x = App10x(app=mock_app)

        # Mock the _run_as_web_server method to capture the callback
        mock_app._run_as_web_server = Mock()

        # Custom callback to test
        callback_called = []

        def test_callback(server):
            callback_called.append(server)

        app10x._run_in_window(on_server_created=test_callback)

        # Get the internal_on_server_created callback
        call_kwargs = mock_app._run_as_web_server.call_args[1]
        internal_callback = call_kwargs['internal_on_server_created']

        # Test the callback
        mock_server = Mock()
        mock_server.config = Mock()
        mock_server.config.app = Mock()
        internal_callback(mock_server)

        # Verify the callback was called
        assert len(callback_called) == 1
        assert callback_called[0] is mock_server

        # Verify server was configured
        assert hasattr(mock_server.config.app, 'app10x')
        assert mock_server.config.app.app10x is app10x


class TestFastapiServer:
    """Test cases for the FastapiServer class."""

    def test_fastapi_server_class_attributes(self):
        """Test FastapiServer class attributes."""
        # Test that FastapiServer has the expected type annotation
        # The app10x is defined as a type annotation, not a class attribute
        # We can check that the class has the annotation in its __annotations__
        assert 'app10x' in FastapiServer.__annotations__
        # The annotation is stored as a string, so we check for the string representation
        assert FastapiServer.__annotations__['app10x'] == 'App10x'

    def test_fastapi_server_inheritance(self):
        """Test that FastapiServer properly inherits from app_server.FastapiServer."""
        from ui_10x.rio.internals.app import app_server

        assert issubclass(FastapiServer, app_server.FastapiServer)

    async def test_fastapi_server_create_session(self):
        """Test create_session method creates Session with app10x attribute."""
        # Create a mock server that behaves like FastapiServer
        server = Mock()
        server.__class__ = FastapiServer
        server.app10x = Mock()

        # Mock the parent create_session method
        mock_session = Mock(spec=rio.Session)
        mock_parent_create_session = Mock()

        async def fake_create_session(*args, **kwargs):
            mock_parent_create_session(*args, **kwargs)
            return mock_session

        with patch.object(app_server.FastapiServer, 'create_session', fake_create_session):
            # Call the actual FastapiServer.create_session method
            result = await FastapiServer.create_session(server)

            # Should call parent method
            mock_parent_create_session.assert_called_once()

            # Should return a Session instance with app10x attribute
            assert result.__class__ == Session
            assert result.app10x is server.app10x


class TestSession:
    """Test cases for the Session class."""

    def test_session_class_attributes(self):
        """Test Session class attributes."""
        # Test that Session has the expected class attributes
        assert hasattr(Session, 'app10x')
        # The app10x attribute should be a class attribute that can be set on instances
        assert isinstance(Session.app10x, type(None)) or hasattr(Session, 'app10x')

    def test_session_inheritance(self):
        """Test that Session properly inherits from rio.Session."""
        assert issubclass(Session, rio.Session)

    def test_session_method_attributes(self):
        """Test that Session has the correct method attributes."""
        # Verify that Session has the expected method attributes
        assert hasattr(Session, '_local_methods_')
        assert hasattr(Session, '_remote_methods_')

        # These should be copies of the parent class attributes
        assert Session._local_methods_ is not rio.Session._local_methods_
        assert Session._remote_methods_ is not rio.Session._remote_methods_
        assert isinstance(Session._local_methods_, dict)
        assert isinstance(Session._remote_methods_, dict)

        # The content should be the same as the parent class
        assert Session._local_methods_ == rio.Session._local_methods_
        assert Session._remote_methods_ == rio.Session._remote_methods_

    def test_session_methods_exist(self):
        """Test that Session has the expected methods."""
        # Test that the custom methods exist on the Session class
        assert hasattr(Session, '_close')
        assert hasattr(Session, '_get_webview_window')
        assert hasattr(Session, 'set_title')
        assert hasattr(Session, 'pick_folder')
        assert hasattr(Session, 'pick_file')
        assert hasattr(Session, 'save_file')

    async def test_session_close_not_running_in_window(self):
        """Test _close when not running in window calls parent method twice."""
        # Create a mock session with the actual Session class behavior
        session = Mock()
        session.__class__ = Session
        session.running_in_window = False

        # Mock the parent _close method
        mock_parent_close = Mock()

        async def fake_close(*args, **kwargs):
            mock_parent_close(*args, **kwargs)

        with patch.object(rio.Session, '_close', fake_close):
            # Call the actual Session._close method
            await Session._close(session, close_remote_session=True)

            # Should call parent method twice: once with original parameter, once with False
            assert mock_parent_close.call_count == 2
            mock_parent_close.assert_any_call(session, close_remote_session=True)
            mock_parent_close.assert_any_call(session, close_remote_session=False)

    async def test_session_close_running_in_window(self):
        """Test _close when running in window calls parent with False and closes webview."""
        session = Mock()
        session.__class__ = Session
        session.running_in_window = True
        session.app10x = Mock()
        session.app10x.webview = Mock()

        # Mock the parent _close method
        mock_parent_close = Mock()

        async def fake_close(*args, **kwargs):
            mock_parent_close(*args, **kwargs)

        with patch.object(rio.Session, '_close', fake_close):
            # Call the actual Session._close method
            await Session._close(session, close_remote_session=True)

            # Should call parent method once with close_remote_session=False
            mock_parent_close.assert_called_once_with(session, close_remote_session=False)

            # Should close webview
            session.app10x.webview.close.assert_called_once()

    async def test_session_get_webview_window_raises_error(self):
        """Test _get_webview_window raises RuntimeError."""
        session = Mock()
        session.__class__ = Session

        with pytest.raises(RuntimeError, match='Should not be called required in out-of-process webview'):
            await Session._get_webview_window(session)

    async def test_session_set_title_not_running_in_window(self):
        """Test set_title when not running in window calls parent method."""
        session = Mock()
        session.__class__ = Session
        session.running_in_window = False

        # Mock the parent set_title method
        mock_parent_set_title = Mock()

        async def fake_set_title(*args, **kwargs):
            mock_parent_set_title(*args, **kwargs)

        with patch.object(rio.Session, 'set_title', fake_set_title):
            # Call the actual Session.set_title method
            await Session.set_title(session, 'Test Title')

            # Should call parent method
            mock_parent_set_title.assert_called_once_with(session, 'Test Title')

    async def test_session_set_title_running_in_window(self):
        """Test set_title when running in window calls webview instead of parent."""
        session = Mock()
        session.__class__ = Session
        session.running_in_window = True
        session.app10x = Mock()
        session.app10x.webview = Mock()

        # Mock the parent set_title method
        mock_parent_set_title = Mock()

        async def fake_set_title(*args, **kwargs):
            mock_parent_set_title(*args, **kwargs)

        with patch.object(rio.Session, 'set_title', fake_set_title):
            # Call the actual Session.set_title method
            await Session.set_title(session, 'Test Title')

            # Should not call parent method
            mock_parent_set_title.assert_not_called()

            # Should call webview set_title
            session.app10x.webview.set_title.assert_called_once_with('Test Title')

    async def test_session_pick_folder_not_running_in_window(self):
        """Test pick_folder when not running in window calls parent method."""
        session = Mock()
        session.__class__ = Session
        session.running_in_window = False
        expected_path = pathlib.Path('/test/path')

        # Mock the parent pick_folder method
        mock_parent_pick_folder = Mock()

        async def fake_pick_folder(*args, **kwargs):
            mock_parent_pick_folder(*args, **kwargs)
            return expected_path

        with patch.object(rio.Session, 'pick_folder', fake_pick_folder):
            # Call the actual Session.pick_folder method
            result = await Session.pick_folder(session)

            # Should call parent method and return its result
            mock_parent_pick_folder.assert_called_once()
            assert result == expected_path

    async def test_session_pick_folder_running_in_window(self):
        """Test pick_folder when running in window calls webview and returns Path."""
        session = Mock()
        session.running_in_window = True
        session.app10x = Mock()
        session.app10x.webview = Mock()
        session.app10x.webview.pick_folder.return_value = '/test/path'

        # Call the actual Session.pick_folder method
        result = await Session.pick_folder(session)

        # Should call webview pick_folder and return Path
        session.app10x.webview.pick_folder.assert_called_once()
        assert result == pathlib.Path('/test/path')

    async def test_session_pick_file_not_running_in_window(self):
        """Test pick_file when not running in window calls parent method."""
        session = Mock()
        session.__class__ = Session
        session.running_in_window = False
        expected_file_info = Mock()

        # Mock the parent pick_file method
        mock_parent_pick_file = Mock()

        async def fake_pick_file(*args, **kwargs):
            mock_parent_pick_file(*args, **kwargs)
            return expected_file_info

        with patch.object(rio.Session, 'pick_file', fake_pick_file):
            # Call the actual Session.pick_file method
            result = await Session.pick_file(session, file_types=['txt'], multiple=False)

            # Should call parent method and return its result
            mock_parent_pick_file.assert_called_once_with(session, file_types=['txt'], multiple=False)
            assert result == expected_file_info

    async def test_session_pick_file_running_in_window_single(self):
        """Test pick_file when running in window (single file) calls webview."""
        session = Mock()
        session.running_in_window = True
        session.app10x = Mock()
        session.app10x.webview = Mock()
        session.app10x.webview.pick_file.return_value = '/test/file.txt'

        # Mock utils.FileInfo._from_path
        mock_file_info = Mock()
        with patch('ui_10x.rio.internals.app.utils.FileInfo._from_path', return_value=mock_file_info):
            # Call the actual Session.pick_file method
            result = await Session.pick_file(session, file_types=['txt'], multiple=False)

            # Should call webview pick_file with normalized file types
            session.app10x.webview.pick_file.assert_called_once_with(file_types=['txt (*.txt)'], multiple=False)
            assert result == mock_file_info

    async def test_session_pick_file_running_in_window_multiple(self):
        """Test pick_file when running in window (multiple files) calls webview."""
        session = Mock()
        session.running_in_window = True
        session.app10x = Mock()
        session.app10x.webview = Mock()
        session.app10x.webview.pick_file.return_value = ['/test/file1.txt', '/test/file2.txt']

        # Mock utils.FileInfo._from_path
        mock_file_info1 = Mock()
        mock_file_info2 = Mock()
        with patch('ui_10x.rio.internals.app.utils.FileInfo._from_path', side_effect=[mock_file_info1, mock_file_info2]):
            # Call the actual Session.pick_file method
            result = await Session.pick_file(session, file_types=['txt'], multiple=True)

            # Should call webview pick_file with normalized file types
            session.app10x.webview.pick_file.assert_called_once_with(file_types=['txt (*.txt)'], multiple=True)
            assert result == [mock_file_info1, mock_file_info2]

    async def test_session_pick_file_no_selection(self):
        """Test pick_file when no file is selected raises NoFileSelectedError."""
        session = Mock()
        session.running_in_window = True
        session.app10x = Mock()
        session.app10x.webview = Mock()
        session.app10x.webview.pick_file.return_value = None

        # Should raise NoFileSelectedError
        with pytest.raises(rio.errors.NoFileSelectedError):
            await Session.pick_file(session, file_types=['txt'], multiple=False)

    async def test_session_pick_file_normalize_file_types(self):
        """Test pick_file normalizes and deduplicates file types."""
        session = Mock()
        session.running_in_window = True
        session.app10x = Mock()
        session.app10x.webview = Mock()
        session.app10x.webview.pick_file.return_value = '/test/file.txt'

        mock_file_info = Mock()
        with patch('ui_10x.rio.internals.app.utils.FileInfo._from_path', return_value=mock_file_info):
            # Call the actual Session.pick_file method
            await Session.pick_file(session, file_types=['txt', 'pdf'], multiple=False)

            # Should call webview with file types formatted correctly
            session.app10x.webview.pick_file.assert_called_once_with(file_types=['txt (*.txt)', 'pdf (*.pdf)'], multiple=False)

    async def test_session_save_file_not_running_in_window(self):
        """Test save_file when not running in window calls parent method."""
        session = Mock()
        session.__class__ = Session
        session.running_in_window = False

        # Mock the parent save_file method
        mock_parent_save_file = Mock()

        async def fake_save_file(*args, **kwargs):
            mock_parent_save_file(*args, **kwargs)

        with patch.object(rio.Session, 'save_file', fake_save_file):
            # Call the actual Session.save_file method
            await Session.save_file(session, 'test content', 'test.txt', media_type='text/plain', directory=pathlib.Path('/test'))

            # Should call parent method
            mock_parent_save_file.assert_called_once_with(
                session, 'test content', 'test.txt', media_type='text/plain', directory=pathlib.Path('/test')
            )

    async def test_session_save_file_running_in_window(self):
        """Test save_file when running in window calls webview."""
        session = Mock()
        session.running_in_window = True
        session.app10x = Mock()
        session.app10x.webview = Mock()

        # Call the actual Session.save_file method
        await Session.save_file(session, 'test content', 'test.txt', media_type='text/plain', directory=pathlib.Path('/test'))

        # Should call webview save_file
        session.app10x.webview.save_file.assert_called_once_with(file_contents='test content', directory=str(pathlib.Path('/test')), file_name='test.txt')

    async def test_session_save_file_running_in_window_no_directory(self):
        """Test save_file when running in window with no directory calls webview."""
        session = Mock()
        session.running_in_window = True
        session.app10x = Mock()
        session.app10x.webview = Mock()

        # Call the actual Session.save_file method
        await Session.save_file(session, 'test content', 'test.txt')

        # Should call webview save_file with empty directory
        session.app10x.webview.save_file.assert_called_once_with(file_contents='test content', directory='', file_name='test.txt')


class TestIntegration:
    """Integration tests for the complete App10x workflow."""

    @patch('ui_10x.rio.internals.app.WebViewProcess')
    @patch('ui_10x.rio.internals.app.utils.ensure_valid_port')
    @patch('asyncio.run')
    def test_complete_app10x_workflow(self, mock_asyncio_run, mock_ensure_port, mock_webview_process):
        """Test the complete App10x workflow from initialization to cleanup."""
        # Setup mocks
        mock_ensure_port.return_value = 8080
        mock_asyncio_run.return_value = '/path/to/icon.png'
        mock_webview = Mock()
        mock_webview.is_alive.return_value = True
        mock_webview_process.return_value = mock_webview

        # Create mock Rio app
        mock_app = Mock()
        mock_app.name = 'Integration Test App'
        mock_app._run_as_web_server = Mock()

        # Create App10x instance
        app10x = App10x(app=mock_app)

        # Test the complete workflow
        app10x._run_in_window(debug_mode=True)

        # Verify WebViewProcess was created
        mock_webview_process.assert_called_once()

        # Verify the server was configured correctly
        mock_app._run_as_web_server.assert_called_once()
        call_kwargs = mock_app._run_as_web_server.call_args[1]
        assert call_kwargs['debug_mode'] is True
        assert call_kwargs['running_in_window'] is True

        # Verify cleanup was performed
        mock_webview.close.assert_called_once()
        mock_webview.join.assert_called_once()

    def test_session_method_attributes(self):
        """Test that Session class has the correct method attributes."""
        # Verify that Session has the expected method attributes
        assert hasattr(Session, '_local_methods_')
        assert hasattr(Session, '_remote_methods_')

        # These should be copies of the parent class attributes
        assert Session._local_methods_ is not rio.Session._local_methods_
        assert Session._remote_methods_ is not rio.Session._remote_methods_
        assert isinstance(Session._local_methods_, dict)
        assert isinstance(Session._remote_methods_, dict)



if __name__ == '__main__':
    #import os;assert os.environ.setdefault('UI_PLATFORM', 'Rio') == os.getenv('UI_PLATFORM')

    from core_10x.code_samples.directories import ANIMALS, FISH

    from ui_10x.choice import Choice
    from ui_10x.utils import UxDialog, ux

    ux.init()

    choice1 = Choice(choices = ANIMALS)
    d = UxDialog(choice1.widget())
    d.exec()

    print(choice1.values_selected)

    choice2 = Choice(choices = FISH)
    d = UxDialog(choice2.widget())
    d.exec()

    print(choice2.values_selected)

    choice3 = Choice(choices = FISH.choices())

    d = UxDialog(choice3.widget())
    d.exec()

    print(choice3.values_selected)

    choice3.popup()
    print(choice3.values_selected)



import asyncio

import rio.testing
from ui_10x.rio.component_builder import DynamicComponent
from ui_10x.rio.widgets import CheckBox


async def test_checkbox_comprehensive() -> None:
    """Test CheckBox with comprehensive client-widget interaction verification."""
    widget = CheckBox('Check me')
    find_checkbox = 'document.querySelector("input[type=\'checkbox\']")'

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # 1) Verify client shows widget value (unchecked initially)
        assert not widget.is_checked()
        client_checked = await test_client.execute_js(find_checkbox + '.checked')
        assert client_checked is False

        # 2) Modify client value (user clicking checkbox)
        await test_client.execute_js(find_checkbox + '.click();')
        await asyncio.sleep(0.5)  # Wait for event processing

        # 3) Verify widget reflects client value
        assert widget.is_checked()
        client_checked = await test_client.execute_js(find_checkbox + '.checked')
        assert client_checked is True

        # Test widget changes propagate to client
        widget.set_checked(False)
        await test_client.wait_for_refresh()
        client_checked = await test_client.execute_js(find_checkbox + '.checked')
        assert client_checked is False
        assert not widget.is_checked()

        # Test another client interaction
        await test_client.execute_js(find_checkbox + '.click();')
        await asyncio.sleep(0.5)
        assert widget.is_checked()


async def test_checkbox_disabled_interaction() -> None:
    """Test CheckBox disabled state blocks user interaction."""
    widget = CheckBox('Test Checkbox')
    find_checkbox = 'document.querySelector("input[type=\'checkbox\']")'

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial enabled state - clicks should work
        await test_client.execute_js(find_checkbox + '.click();')
        await asyncio.sleep(0.5)
        assert widget.is_checked()

        # Disable widget
        widget.set_enabled(False)
        await test_client.wait_for_refresh()

        # Verify client shows disabled state
        client_disabled = await test_client.execute_js(find_checkbox + '.disabled')
        assert client_disabled is True

        # Test that clicks are blocked when disabled
        initial_checked = widget.is_checked()
        await test_client.execute_js(find_checkbox + '.click();')
        await asyncio.sleep(0.5)
        assert widget.is_checked() == initial_checked  # State should not change

        # Re-enable widget
        widget.set_enabled(True)
        await test_client.wait_for_refresh()

        # Verify client shows enabled state
        client_disabled = await test_client.execute_js(find_checkbox + '.disabled')
        assert client_disabled is False

        # Test that clicks work again
        await test_client.execute_js(find_checkbox + '.click();')
        await asyncio.sleep(0.5)
        assert not widget.is_checked()  # Should toggle


async def test_checkbox_basic_functionality() -> None:
    """Test CheckBox basic functionality without client interaction."""
    widget = CheckBox('Test Checkbox')

    async with rio.testing.BrowserClient(lambda: DynamicComponent(widget)) as test_client:
        await asyncio.sleep(0.5)

        # Test initial state
        assert not widget.is_checked()

        # Test checking
        widget.set_checked(True)
        await test_client.wait_for_refresh()
        assert widget.is_checked()

        # Test unchecking
        widget.set_checked(False)
        await test_client.wait_for_refresh()
        assert not widget.is_checked()

        # Test enabled/disabled state
        widget.set_enabled(False)
        await test_client.wait_for_refresh()

        widget.set_enabled(True)
        await test_client.wait_for_refresh()




if __name__ == '__main__':
    from datetime import date

    from core_10x.environment_variables import EnvVars, XDateTime

    d = date.today()
    print(XDateTime.date_to_str(d))

    df_10x = EnvVars.date_format
    print(XDateTime.date_to_str(d))




from core_10x.traitable import Traitable, T, RT

class Book(Traitable):
    name: str       = T(T.ID)
    anything: bool  = T()

if __name__ == '__main__':
    from core_10x.manual_tests.existing_instance_bug import Book

    b1 = Book(name = 'B1')
    b2 = Book(_replace = True, name = 'B2', anything = False)

    try:
        b11 = Book.existing_instance(name = 'B1')
    except ValueError as ex:
        print(str(ex))

    b21 = Book.existing_instance(name = 'B2')




from __future__ import annotations

import copy
from collections import defaultdict

from py10x_kernel import BFlags, BTraitFlags
from typing_extensions import Never

from core_10x.ui_hint import Ui, UiHintModification
from core_10x.xnone import XNone, XNoneType

# fmt: off
#---- Attribute Tags
NAME_TAG        = 'name'
DATATYPE_TAG    = 'data_type'
FLAGS_TAG       = 'flags'
DEFAULT_TAG     = 'default'
FORMAT_TAG      = 'fmt'
PARAMS_TAG      = 'params'
GETTER_PARAMS_TAG = 'getter_params'
UI_HINT_TAG     = 'ui_hint'
# fmt: on


class TraitDefinition:
    # fmt: off
    __slots__ = (
        NAME_TAG,
        DATATYPE_TAG,
        FLAGS_TAG,
        DEFAULT_TAG,
        FORMAT_TAG,
        PARAMS_TAG,
        UI_HINT_TAG,
    )

    s_known_attributes = {
        DATATYPE_TAG:   lambda: XNoneType,
        FLAGS_TAG:      lambda: BFlags(0x0),
        DEFAULT_TAG:    lambda: XNone,
        FORMAT_TAG:     lambda: '',
        UI_HINT_TAG:    lambda: Ui(),
    }
    # fmt: on

    def __init__(self, *args, **kwargs):
        """
        T([default_value,] [t_flags,] [Ui(...),] **kwargs
        """
        self.process_args(args, kwargs)

        for tag, def_value_fn in self.s_known_attributes.items():
            def_value = def_value_fn()
            value = kwargs.pop(tag, def_value)
            setattr(self, tag, value)

        self.params = kwargs
        self.name = None

    def __floordiv__(self, comment):
        """Add or replace comment in params with // operator"""
        assert isinstance(comment, str), f'Trait comment must be a string: {comment}'
        ui_hint = getattr(self, UI_HINT_TAG)
        ui_hint.tip = comment
        return self

    def __call__(self, *args, **kwargs):
        """Prevent IDE from complaining about calling traits with parameters"""
        ...

    def set_widget_type(self, widget_type: Ui.WIDGET_TYPE):
        ui_hint = getattr(self, UI_HINT_TAG)
        ui_hint.widget_type = widget_type

    def copy(self) -> TraitDefinition:
        return copy.deepcopy(self)

    def flags_change(self, flags_value):
        dt = type(flags_value)
        if dt is tuple:
            assert len(flags_value) == 2, f'Changing {self.name} flags expects (flags_to_set, flags_to_reset)'
            self.flags.set_reset(*(0 if flag is None else flag.value() for flag in flags_value))
        else:
            self.flags.set(flags_value.value())

    @classmethod
    def process_args(cls, args: tuple, kwargs: dict):
        """
        args:
            optional default_value
            optional T.flags
            optional ui hint: Ui(...)
        kwargs:
            known attributes and extra params (usually empty)
        """
        the_args = list(args)
        already_processed = defaultdict(bool)

        n = len(the_args)
        while n:
            arg = the_args.pop(0)
            dt = type(arg)
            if dt is Ui:  # -- only Ui hint is given, no more args
                assert not already_processed[UI_HINT_TAG], f'Duplicated Ui hint: {args}'
                assert n == 1, f'No positional args after Ui hint: {args}'
                kwargs[UI_HINT_TAG] = arg
                already_processed[UI_HINT_TAG] = True

            elif issubclass(dt, BFlags):  # -- no default value => n <=2
                assert not already_processed[FLAGS_TAG], f'Duplicated T.flags: {args}'
                assert n <= 2, f'Too many positional args after T.flags: {args}'
                kwargs[FLAGS_TAG] = arg
                already_processed[FLAGS_TAG] = True

            else:  # -- must be a default value
                assert not already_processed[DEFAULT_TAG], f'Duplicated default value: {args}'
                kwargs[DEFAULT_TAG] = arg
                already_processed[DEFAULT_TAG] = True

            n -= 1


class TraitModification(TraitDefinition):
    # fmt: off
    s_known_attributes = {
        FLAGS_TAG:      lambda: Never,
        DEFAULT_TAG:    lambda: Never,
        FORMAT_TAG:     lambda: Never,
        UI_HINT_TAG:    UiHintModification
    }

    s_modifiers = {
        FLAGS_TAG:      lambda self, flags_modification: self.flags_change(flags_modification),
        UI_HINT_TAG:    lambda self, ui_hint_modification: setattr(self,UI_HINT_TAG,ui_hint_modification.apply(self.ui_hint))
    }
    # fmt: on
    def apply(self, trait_def: TraitDefinition) -> TraitDefinition:
        res = trait_def.copy()
        modifiers = self.__class__.s_modifiers
        for attr_name in self.__class__.s_known_attributes:
            modified_value = getattr(self, attr_name)
            if modified_value is not Never:
                modifier = modifiers.get(attr_name)
                if modifier:
                    modifier(res, modified_value)
                else:
                    setattr(res, attr_name, modified_value)
        res.params.update(self.params)
        return res


class T(BTraitFlags):
    def __new__(cls, *args, **kwargs) -> TraitDefinition:
        return TraitDefinition(*args, **kwargs)

    @staticmethod
    def fg_color(color: str) -> str:
        return f'color: {color}' if color else ''

    @staticmethod
    def bg_color(color: str) -> str:
        return f'background-color: {color}' if color else ''

    @staticmethod
    def colors(bg_color: str, fg_color: str) -> str:
        return f'background-color: {bg_color}; color: {fg_color}' if bg_color and fg_color else ''

    STICKY = BTraitFlags.OFFGRAPH_SET

def RT(*args, **kwargs) -> TraitDefinition:  # noqa: N802
    trait_def = TraitDefinition(*args, **kwargs)
    trait_def.flags.set(T.RUNTIME.value())
    return trait_def


def M(*args, **kwargs) -> TraitModification:  # noqa: N802
    return TraitModification(*args, **kwargs)



from core_10x.environment_variables import EnvVars
from core_10x.global_cache import cache
from core_10x.py_class import PyClass


# ===================================================================================================================================
#   Data Domain Bindings
#
#   from core_10x.data_domain import GeneralDomain
#   from xxx.yyy.mdu_domain import MDU
#   ...
#   DATA_DOMAIN_BINDINGS = dict(
#       dev = {
#           GeneralDomain: dict(
#               GENERAL             = R(TS_STORE.MONGO_DB,          hostname = 'dev.mongo.general.io', tsl = True, ...),
#           ),
#
#           MDU:    dict(
#               SYMBOLOGY           = R(REL_DB.ORACLE_DB,           hostname = 'dev.oracle.io', a = '...', b = '...'),
#               MKT_CLOSE_CLUSTER   = R(CLOUD_CLUSTER.RAY_CLUSTER,  hostname = 'dev.ray1.io', ...),
#           ),
#       },
#
#       prod = {
#           GeneralDomain: dict(
#               GENERAL             = R(TS_STORE.MONGO_DB,          hostname = 'prod.mongo.general.io', tsl = True, ...),
#           ),
#
#           MDU:    dict(
#               SYMBOLOGY           = R(REL_DB.ORACLE_DB,           hostname = 'prod.oracle.io', a = '...', b = '...'),
#               MKT_CLOSE_CLUSTER   = R(CLOUD_CLUSTER.RAY_CLUSTER,  hostname = 'prod.ray2.io', ...),
#           ),
#       },
#
#   )
# ===================================================================================================================================
class DataDomainBinder:
    @staticmethod
    @cache
    def all_bindings() -> dict:
        dd_bindings_symbol = EnvVars.data_domain_bindings
        if not dd_bindings_symbol:
            return None

        return PyClass.find_symbol(dd_bindings_symbol)



import os

import pytest
import rio.testing.browser_client


def running_with_coverage(config):
    if not config.pluginmanager.getplugin('pytest_cov'):
        return False

    if not config.getoption('--cov', default='COV_CORE_SOURCE' in os.environ):
        return False

    return not config.getoption('--no-cov', default=False)


@pytest.fixture(scope='session', autouse=True)
async def manage_server(request):
    if running_with_coverage(request.config):
        # run headless client even if running with coverage
        rio.testing.browser_client.DEBUGGER_ACTIVE = False
        pytest.mark.timeout(180)
    else:
        if rio.testing.browser_client.DEBUGGER_ACTIVE:
            pytest.mark.timeout(0 if rio.testing.browser_client.DEBUGGER_ACTIVE else 90)

    async with rio.testing.browser_client.prepare_browser_client():
        yield


@pytest.fixture(autouse=True)
def setup_ui_platform(monkeypatch):
    monkeypatch.setenv('UI_PLATFORM', 'Rio')
    yield
    monkeypatch.undo()



from __future__ import annotations

from py10x_kernel import BNucleus


class Nucleus:
    __slots__           = ()
    TYPE_TAG            = BNucleus.TYPE_TAG
    CLASS_TAG           = BNucleus.CLASS_TAG
    REVISION_TAG        = BNucleus.REVISION_TAG
    OBJECT_TAG          = BNucleus.OBJECT_TAG
    COLLECTION_TAG      = BNucleus.COLLECTION_TAG
    ID_TAG              = BNucleus.ID_TAG
    NX_RECORD_TAG       = BNucleus.NX_RECORD_TAG
    TYPE_RECORD_TAG     = BNucleus.TYPE_RECORD_TAG
    PICKLE_RECORD_TAG   = BNucleus.PICKLE_RECORD_TAG

    serialize_any       = BNucleus.serialize_any
    deserialize_any     = BNucleus.deserialize_any

    serialize_type      = BNucleus.serialize_type
    deserialize_type    = BNucleus.deserialize_type
    serialize_complex   = BNucleus.serialize_complex
    deserialize_complex = BNucleus.deserialize_complex
    serialize_date      = BNucleus.serialize_date
    deserialize_date    = BNucleus.deserialize_date
    serialize_list      = BNucleus.serialize_list
    deserialize_list    = BNucleus.deserialize_list
    serialize_dict      = BNucleus.serialize_dict
    deserialize_dict    = BNucleus.deserialize_dict
    deserialize_record  = BNucleus.deserialize_record

    #===================================================================================================================
    #   The following methods must be implemented by a subclass of Nucleus
    #===================================================================================================================

    def __repr__(self):
        return self.to_str()

    def __eq__(self, other):
        return self.__class__.same_values(self, other)

    def to_str(self) -> str:
        return str(self)

    def to_id(self) -> str:
        return self.to_str()

    def serialize(self, embed: bool):
        raise NotImplementedError

    @classmethod
    def deserialize(cls, serialized_data) -> Nucleus:
        raise NotImplementedError

    @classmethod
    def from_str(cls, s: str) -> Nucleus:
        raise NotImplementedError

    @classmethod
    def from_any_xstr(cls, value) -> Nucleus:
        raise NotImplementedError

    @classmethod
    def from_any(cls, value) -> Nucleus:
        if isinstance(value, cls):
            return value

        if isinstance(value, str):
            return cls.from_str(value)

        return cls.from_any_xstr(value)

    @classmethod
    def same_values(cls, value1, value2) -> bool:
        raise NotImplementedError

    @classmethod
    def choose_from(cls):
        return {}



from __future__ import annotations

from datetime import date  # noqa: TC003

import pytest
from core_10x.trait_filter import (
    AND,
    BETWEEN,
    EQ,
    GE,
    GT,
    IN,
    LE,
    LT,
    NE,
    NIN,
    NOT_EMPTY,
    OR,
    f,
)
from core_10x.traitable import Traitable, XNone


class Person(Traitable):
    first_name: str
    last_name: str
    age: int
    dob: date


def test_filters():
    p = Person(first_name='Sasha', last_name='Davidovich')

    r = OR(f(age=BETWEEN(50, 70), first_name=NE('Sasha')), f(age=17))

    assert r.prefix_notation() == {'$or': [{'age': {'$gte': 50, '$lte': 70}, 'first_name': {'$ne': 'Sasha'}}, {'age': {'$eq': 17}}]}
    assert not r.eval(p)


def test_or():
    p = Person(first_name='Sasha', last_name='Davidovich')
    r1, r2 = f(age=BETWEEN(50, 70)), f(first_name=EQ('Sasha'))
    r3 = OR(r1, r2)
    r4 = OR(r1)
    r5 = OR(r2)
    r6 = OR()
    assert r1.prefix_notation() == {'age': {'$gte': 50, '$lte': 70}}
    assert r2.prefix_notation() == {'first_name': {'$eq': 'Sasha'}}
    assert r3.prefix_notation() == {'$or': [r1.prefix_notation(), r2.prefix_notation()]}

    assert r4.prefix_notation() == r1.prefix_notation()
    assert r5.prefix_notation() == r2.prefix_notation()
    assert r6.prefix_notation() == {'$in': []}

    assert not r1.eval(p)
    assert r2.eval(p)
    assert r3.eval(p)
    assert not r4.eval(p)
    assert r5.eval(p)
    assert not r6.eval(p)

    r7 = OR(OR(), OR())
    assert r7.prefix_notation() == {'$in': []}
    assert not r7.eval(p)

    r8 = OR(OR(), AND())
    assert r7.prefix_notation() == {'$in': []}
    assert r8.eval(p)


def test_and():
    p = Person(first_name='Sasha', last_name='Davidovich')

    r1, r2 = f(age=BETWEEN(50, 70)), f(first_name=EQ('Sasha'))
    r3 = AND(r1, r2)
    r4 = AND(r1)
    r5 = AND(r2)
    r6 = AND()
    assert r1.prefix_notation() == {'age': {'$gte': 50, '$lte': 70}}
    assert r2.prefix_notation() == {'first_name': {'$eq': 'Sasha'}}
    assert r3.prefix_notation() == {'$and': [r1.prefix_notation(), r2.prefix_notation()]}

    assert r4.prefix_notation() == r1.prefix_notation()
    assert r5.prefix_notation() == r2.prefix_notation()
    assert r6.prefix_notation() == {}

    assert not r1.eval(p)
    assert r2.eval(p)
    assert not r3.eval(p)
    assert not r4.eval(p)
    assert r5.eval(p)
    assert r6.eval(p)

    r7 = AND(AND(), AND())
    assert r7.prefix_notation() == {}
    assert r7.eval(p)

    r8 = AND(AND(), OR())
    assert r8.prefix_notation() == {'$in': []}
    assert not r8.eval(p)


def test_simple_ops_eval_and_prefix():
    assert EQ(5).eval(5)
    assert not EQ(5).eval(4)
    assert EQ('x').prefix_notation() == {'$eq': 'x'}

    assert NE(5).eval(4)
    assert not NE(5).eval(5)
    assert NE('x').prefix_notation() == {'$ne': 'x'}

    assert GT(5).eval(6)
    assert not GT(5).eval(5)
    assert GT(5).prefix_notation() == {'$gt': 5}

    assert GE(5).eval(5)
    assert GE(5).eval(6)
    assert not GE(5).eval(4)
    assert GE(5).prefix_notation() == {'$gte': 5}

    assert LT(5).eval(4)
    assert not LT(5).eval(5)
    assert LT(5).prefix_notation() == {'$lt': 5}

    assert LE(5).eval(5)
    assert LE(5).eval(4)
    assert not LE(5).eval(6)
    assert LE(5).prefix_notation() == {'$lte': 5}


def test_in_nin():
    assert IN([1, 2, 3]).eval(2)
    assert not IN([1, 2, 3]).eval(4)
    assert IN((1, 2)).prefix_notation() == {'$in': (1, 2)}

    assert NIN([1, 2, 3]).eval(4)
    assert not NIN([1, 2, 3]).eval(2)
    assert NIN([1, 2]).prefix_notation() == {'$nin': [1, 2]}


def test_between_bounds_and_prefix():
    b = BETWEEN(1, 5)
    assert b.eval(3)
    assert b.eval(1)
    assert b.eval(5)
    assert not b.eval(0)
    assert not b.eval(6)
    assert b.prefix_notation() == {'$gte': 1, '$lte': 5}

    b_ex = BETWEEN(1, 5, bounds=(False, False))
    assert b_ex.eval(2)
    assert not b_ex.eval(1)
    assert not b_ex.eval(5)
    assert b_ex.prefix_notation() == {'$gt': 1, '$lt': 5}


def test_not_empty_and_bool_ops_eval_and_prefix():
    assert NOT_EMPTY().eval('abc')
    assert not NOT_EMPTY().eval('')
    with pytest.raises(NotImplementedError):
        NOT_EMPTY().prefix_notation()

    a = AND(EQ(5), GT(3))
    assert a.eval(5)
    assert not a.eval(3)
    assert a.prefix_notation() == {'$and': [{'$eq': 5}, {'$gt': 3}]}

    o = OR(EQ(1), EQ(2))
    assert o.eval(1)
    assert o.eval(2)
    assert not o.eval(3)
    assert o.prefix_notation() == {'$or': [{'$eq': 1}, {'$eq': 2}]}

    # single-argument behaviors: prefix_notation returns inner dict
    single_and = AND(EQ(7))
    assert single_and.prefix_notation() == {'$eq': 7} or single_and.prefix_notation() == {'$eq': 7}

    # empty BoolOp
    assert OR().prefix_notation() == {'$in': []}
    assert AND().prefix_notation() == {}


def test_f_named_expressions_eval_and_prefix():
    d = Person(age=10, first_name='Bob', last_name='')

    # named expressions may be raw values (wrapped to EQ) or filters
    filt = f(age=EQ(10), first_name='Bob')
    assert filt.eval(d)
    assert filt.prefix_notation() == {'age': {'$eq': 10}, 'first_name': {'$eq': 'Bob'}}

    # mismatch
    filt2 = f(age=EQ(11), first_name='Bob')
    assert not filt2.eval(d)

    # ensure f uses Person.get_value for named fields
    assert f(age=10).eval(d)  # 10 wrapped as EQ(10) -> matches d.get_value('age') == 10
    assert not f(age=9).eval(d)

    # f with multiple named expressions
    multi = f(age=EQ(10), first_name=NE('Alice'), last_name=NOT_EMPTY())
    assert multi.eval(Person(age=10, first_name='Bob', last_name='Smith'))
    assert not f(last_name=NOT_EMPTY()).eval(Person(last_name=''))  # empty string -> NOT_EMPTY false


def test_named_serializers():
    class P(Person):
        @classmethod
        def age_serialize(cls, t, v):
            return f'age:{v}'  # noinspection PyUnusedLocal

    trait = P.trait('age')

    assert trait is P.trait('age')

    assert trait.serialize_value(5) == 'age:5'

    assert EQ(5).prefix_notation(trait=trait, traitable_class=P.s_bclass) == {'$eq': 'age:5'}

    assert BETWEEN(1, 5).prefix_notation(trait=trait, traitable_class=P.s_bclass) == {
        '$gte': 'age:1',
        '$lte': 'age:5',
    }

    x = OR(f(age=LE(70)), f(first_name=NE('Sasha')), f(last_name=XNone))
    assert x.prefix_notation(traitable_class=P.s_bclass) == {
        '$or': [{'age': {'$lte': 'age:70'}}, {'first_name': {'$ne': 'Sasha'}}, {'last_name': {'$eq': None}}]
    }

    x = f(age=BETWEEN(50, 70), first_name=NE('Sasha'))

    assert f(x, P.s_bclass).prefix_notation() == x.prefix_notation(traitable_class=P.s_bclass)

    r = OR(f(age=BETWEEN(50, 70), first_name=NE('Sasha')), f(age=17))
    assert r.prefix_notation(traitable_class=P.s_bclass) == {
        '$or': [
            {'age': {'$gte': 'age:50', '$lte': 'age:70'}, 'first_name': {'$ne': 'Sasha'}},
            {'age': {'$eq': 'age:17'}},
        ]
    }



from __future__ import annotations

from typing import Literal

import rio


class RioTreeItem(rio.Component):
    """same as SimpleTreeItem, but includes tooltip and supports double-click"""

    text: str = ''
    on_double_press: rio.EventHandler[[]] = None
    on_press: rio.EventHandler[[]] = None
    on_change: rio.EventHandler[[]] = None
    tooltip: str | None = None
    editable: bool = False
    editing: bool = False
    children: list[RioTreeItem] = []
    is_expanded: bool = False

    def build_primary_text(self):
        if not self.editing:
            return rio.Text(self.text, justify='left', selectable=False)
        return rio.TextInput(
            self.text,
            align_x=0,  # justify left?
            on_confirm=self.handle_edit_confirm,
        )

    def build_content(self):
        content = self.build_primary_text()
        if self.tooltip:
            content = rio.Row(content, rio.Tooltip(anchor=content, tip=self.tooltip))
        if self.on_double_press:
            content = rio.PointerEventListener(content, on_double_press=self.handle_double_press)
        return content

    def handle_double_press(self, ev: rio.PointerEvent):
        if self.editable:
            self.editing = True
        if self.on_double_press:
            self.on_double_press()

    def handle_edit_confirm(self, text):
        assert self.editing
        self.text = text
        if self.on_change:
            self.on_change()

    def build(self):
        return rio.SimpleTreeItem(
            content=self.build_content(),
            children=[child.build() for child in self.children],
            is_expanded=self.is_expanded,
            on_press=self.on_press,
        )


class RioTreeView(rio.Component):
    """makes item-level callbacks available on the tree level"""

    children: list[rio.Component] = None
    selection_mode: Literal['none', 'single', 'multiple'] = ('none',)
    col_count: int = 1
    header_labels: list[str] | None = None

    def __post__init__(self):
        if self.header_labels is None:
            self.header_labels = ['header']

    def build(self):
        return rio.TreeView(
            *[item.build() for item in self.children],  # TODO: should not need to build each item
            selection_mode=self.selection_mode,
        )



from __future__ import annotations

from math import exp, log

from core_10x.named_constant import (
    Enum,
    EnumBits,
    ErrorCode,
    NamedConstant,
    NamedConstantTable,
    NamedConstantValue,
    Nucleus,
)


class COLOR(NamedConstant, lowercase_values = True):
    RED         = ()
    BLUE        = ()
    GREEN       = ()
    LIGHTGREEN  = ()

class XCOLOR(COLOR):
    BLACK       = ()
    WHITE       = ()

class WEIGHT(NamedConstant):
    LB  = ()
    KG  = ()
    CT  = ()

class COMPOUNDING(NamedConstant):
    SIMPLE      = ()
    ANNUAL      = ()
    SEMI_ANNUAL = ()
    QUARTERLY   = ()
    MONTHLY     = ()
    WEEKLY      = ()
    CONTINUOUS  = ()

class COMPOUND_TRANSFORM(NamedConstant):
    RATE_TO_ACCRUAL = ()
    ACCRUAL_TO_RATE = ()


COLOR_WEIGHT = NamedConstantValue(
    COLOR,
    RED         = 10,
    BLUE        = 20,
    GREEN       = 30,
    LIGHTGREEN  = 35
)

COLOR_CODE = NamedConstantValue(
    COLOR,
    RED         = 1.0,
    BLUE        = 11.1,
    GREEN       = 12.2,
    LIGHTGREEN  = 13.3
)

COMPOUNDING_TRANSFORM_TABLE = NamedConstantTable(COMPOUNDING, COMPOUND_TRANSFORM,
   #               RATE_TO_ACCRUAL                             ACCRUAL_TO_RATE
   SIMPLE      = ( lambda t, r:  1. + r * t,                   lambda t, a: (a - 1.) / t                       if t else 0. ),
   ANNUAL      = ( lambda t, r: (1. + r      ) **  t,          lambda t, a:  a ** (1. / t)       - 1.          if t else 0. ),
   SEMI_ANNUAL = ( lambda t, r: (1. + r /  2.) ** (t *  2.),   lambda t, a: (a ** (1. / t /  2.) - 1.) * 2.    if t else 0. ),
   QUARTERLY   = ( lambda t, r: (1. + r /  4.) ** (t *  4.),   lambda t, a: (a ** (1. / t /  4.) - 1.) * 4.    if t else 0. ),
   MONTHLY     = ( lambda t, r: (1. + r / 12.) ** (t * 12.),   lambda t, a: (a ** (1. / t / 12.) - 1.) * 12.   if t else 0. ),
   WEEKLY      = ( lambda t, r: (1. + r / 52.) ** (t * 52.),   lambda t, a: (a ** (1. / t / 52.) - 1.) * 52.   if t else 0. ),
   CONTINUOUS  = ( lambda t, r: exp(r * t),                    lambda t, a: log(a) / t ),
)

class EVEN_NUMBER(Enum, step = 2):
    ZERO    = ()
    TWO     = ()
    FOUR    = ()

class A_PROBLEM(ErrorCode):
    DOES_NOT_EXIST  = 'Entity {cls}.{id} does not exist'
    REV_CONFLICT    = 'Entity {cls}.{id} - revision {rev} is outdated'
    SAVE_FAILED     = 'Failed to save entity {cls}.{id}'

class ICON(EnumBits):
    FILE = ('material/file',)
    STOP = ('material/stop',)

class TEXT_ALIGN(NamedConstant):
    s_vertical = 0xf << 4

    LEFT = 1
    CENTER = 6
    RIGHT = 11
    TOP = LEFT << 4
    V_CENTER = CENTER << 4
    BOTTOM = RIGHT << 4

    @classmethod
    def from_str(cls, s: str) -> TEXT_ALIGN:
        return super().from_str(s.upper()) # type: ignore[return-value]

    def rio_attr(self) -> str:
        return 'align_y' if self.value & self.s_vertical else 'align_x'

    def rio_value(self) -> float:
        return ((self.value >> 4 if self.value & self.s_vertical else self.value)-1) /10

class FREQUENCY(NamedConstant):
    CALDAY  = ()
    WEEK    = ()

class PARAMS(NamedConstant):
    CHAR        = ()
    FUNC        = ()
    MIN_FREQ    = ()

from dateutil.relativedelta import relativedelta as delta

FREQ_TABLE = NamedConstantTable(FREQUENCY, PARAMS,
    #           CHAR    FUNC                                MIN_FREQ
    CALDAY  =   ('C',   lambda d, n: d + delta(days = n),   FREQUENCY.CALDAY ),
    WEEK    =   ('W',   lambda d, n: d + delta(weeks = n),  FREQUENCY.CALDAY),
)

class XFREQUENCY(FREQUENCY):
    MONTH   = ()
    QUARTER = ()

XFREQ_TABLE = FREQ_TABLE.extend(XFREQUENCY,
    #           CHAR    FUNC                                    MIN_FREQ
    MONTH   =   ('M',   lambda d, n: d + delta(months = n),     XFREQUENCY.MONTH),
    QUARTER =   ('Q',   lambda d, n: d + delta(months = n*3),   XFREQUENCY.MONTH),
)

from core_10x.traitable import Traitable, T

class TestNamedConstant(Traitable):
    count: int      = T()
    freq: FREQUENCY = T()

if __name__ == '__main__':
    from core_10x.manual_tests.named_constant_test import COLOR, XCOLOR, XFREQUENCY

    print( COLOR.LIGHTGREEN.value)
    sgreen = Nucleus.serialize_any(COLOR.GREEN, False)
    swhite = Nucleus.serialize_any(XCOLOR.WHITE, False)
    print(sgreen)
    print(swhite)

    assert Nucleus.deserialize_dict(swhite) is XCOLOR.WHITE
    #print(COLOR_WEIGHT.GREEN, COLOR_CODE.GREEN) #TODO: FIX!

    print(A_PROBLEM.REV_CONFLICT(cls = 'XXX', id = '145678', rev = '3'))

    x,y = TEXT_ALIGN.from_str('left'), TEXT_ALIGN.from_str('v_center')
    print( x, y, x.rio_attr(), y.rio_attr(), x.value, y.value, x.rio_value(), y.rio_value() )

    t1 = TestNamedConstant(count = 2, freq = FREQUENCY.WEEK)
    s1 = t1.serialize_object()

    t2 = TestNamedConstant(count = 2, freq = XFREQUENCY.MONTH)
    s2 = t2.serialize_object()

    t22 = TestNamedConstant.deserialize(s2)
    assert t2.freq == t22.freq


from __future__ import annotations

import importlib
import inspect

from core_10x.global_cache import cache
from core_10x.py_class import PyClass

CLASS_ID_DELIMITER = '/'


class PackageRefactoring:
    # fmt: off
    s_module_name   = '_refactoring'
    s_records_name  = '_records'
    # fmt: on
    s_instances = {}

    @classmethod
    def register_top_level_packages(cls, *top_level_packages):
        instances = cls.s_instances
        for top_level_package in top_level_packages:
            pkg_refact = cls(top_level_package)
            instances[top_level_package] = pkg_refact

    @staticmethod
    def default_class_id(cls: type = None, canonical_class_name: str = None) -> str:
        if canonical_class_name is None:
            assert cls and inspect.isclass(cls), f'{cls} is not a valid class'
            canonical_class_name = PyClass.name(cls)
            assert '__main__' not in canonical_class_name, f'{cls} is defined in __main__ and does not have a stable id'
            assert '<locals>' not in canonical_class_name, f'{cls} is defined locally in a function and cannot be reconstructed'
        return canonical_class_name.replace('.', CLASS_ID_DELIMITER)

    def __init__(self, top_package_name: str):
        try:
            pkg_init = importlib.import_module(top_package_name)
            self.file_name = pkg_init.__file__.replace('__init__', self.__class__.s_module_name)
        except Exception as e:
            raise OSError(f'Unknown top-level package {top_package_name}') from e

        self.package_name = top_package_name
        try:
            module = importlib.import_module(self.s_module_name, top_package_name)
        except Exception:
            self.cid_to_path = {}
            self.path_to_cid = {}
            return

        self.cid_to_path = records = getattr(module, self.s_records_name, None)
        if records is None:
            raise OSError(f'{top_package_name}.{self.s_module_name} is missing refactoring records {self.s_records_name}')

        self.path_to_cid = path_to_cid = {}
        for class_id, canonical_class_name in records.items():
            assert isinstance(class_id, str), f'{class_id} must be a str'
            assert isinstance(canonical_class_name, str), f'{canonical_class_name} must be a canonical class name'

            path_to_cid[canonical_class_name] = class_id

    def _refactor(self, cur_path: str, new_path: str):
        cid = self.path_to_cid.get(cur_path)
        if cid is None:
            raise OSError(f'Unknown canonical class name {cur_path}')

        self.cid_to_path[cid] = new_path
        del self.path_to_cid[cur_path]
        self.path_to_cid[new_path] = cid

    def _save(self):
        text = [f'{self.__class__.s_records_name} = {{']
        for cid, path in self.cid_to_path.items():
            text.append(f'\t{cid}:\t\t"{path}"')
        text.append('}')
        s = '\n'.join(text)
        with open(self.file_name, mode='w') as f:
            f.write(s)

    @staticmethod
    @cache
    def find_class_id(cls) -> str:
        canonical_class_name = PyClass.name(cls)
        pkg_refact: PackageRefactoring
        for pkg_refact in PackageRefactoring.s_instances.values():
            cid = pkg_refact.path_to_cid.get(canonical_class_name)
            if cid is not None:
                return cid

        return PackageRefactoring.default_class_id(cls)

    @staticmethod
    @cache
    def find_class(class_id: str):
        pkg_refact: PackageRefactoring
        for pkg_refact in PackageRefactoring.s_instances.values():
            path = pkg_refact.cid_to_path.get(class_id)
            if path is not None:
                canonical_class_name = path
                break
        else:
            canonical_class_name = class_id.replace(CLASS_ID_DELIMITER, '.')

        cls = PyClass.find(canonical_class_name)
        if cls is None or not inspect.isclass(cls):
            raise OSError(f'Unknown class {canonical_class_name}; class_id = {class_id}')

        return cls

    @staticmethod
    def _find_or_create_instance(canonical_class_name) -> PackageRefactoring:
        parts = canonical_class_name.split('.', maxsplit=1)
        if len(parts) < 2:
            raise ValueError(f'Invalid canonical class name {canonical_class_name}')

        package_name = parts[0]
        pkg_refactor: PackageRefactoring = PackageRefactoring.s_instances.get(package_name)
        if not pkg_refactor:
            pkg_refactor = PackageRefactoring(package_name)
            PackageRefactoring.s_instances[package_name] = pkg_refactor

        return pkg_refactor

    @staticmethod
    def remove_class(canonical_class_name: str, save: bool = True):
        pkg_refactor = PackageRefactoring._find_or_create_instance(canonical_class_name)
        cid = pkg_refactor.path_to_cid.get(canonical_class_name)
        if cid is None:
            raise ValueError(f'Unknown canonical class name {canonical_class_name}')

        del pkg_refactor.cid_to_path[canonical_class_name]
        del pkg_refactor.path_to_cid[cid]

        if save:
            pkg_refactor._save()

    @staticmethod
    def move_class(cur_path: str, new_path: str):
        refactor1 = PackageRefactoring._find_or_create_instance(cur_path)
        refactor2 = PackageRefactoring._find_or_create_instance(new_path)
        the_same = refactor1 is refactor2
        cid = refactor1.path_to_cid.get(cur_path)
        if cid is not None:
            del refactor1.path_to_cid[cur_path]
            del refactor1.cid_to_path[cid]
            if not the_same:
                refactor1._save()

        else:
            cid = PackageRefactoring.default_class_id(canonical_class_name=cur_path)

        refactor2.cid_to_path[cid] = new_path
        refactor2.path_to_cid[new_path] = cid
        refactor2._save()



from __future__ import annotations

import math
from datetime import date

import pytest
from xx_common.curve import IP_KIND, Curve, CurveParams, DateCurve, TwoFuncInterpolator


class TestCurve:
    def test_update_inserts_sorted_and_overwrites(self):
        c = Curve()

        # Insert out of order to check internal sorting
        c.update(5, 50.0, reset=False)
        c.update(1, 10.0, reset=False)
        c.update(3, 30.0, reset=False)

        # Overwrite existing point
        c.update(3, 35.0, reset=False)

        assert c.times == [1, 3, 5]
        assert c.values == [10.0, 35.0, 50.0]

    def test_value_linear_interpolation(self):
        c = Curve()
        c.update_many([0.0, 1.0], [0.0, 2.0], reset=True)

        # Ensure interpolation is allowed for all t by setting a non-None beginning_of_time
        c.beginning_of_time = 0.0

        # Exact nodes
        assert c.value(0.0) == pytest.approx(0.0)
        assert c.value(1.0) == pytest.approx(2.0)

        # Linear interpolation between nodes
        assert c.value(0.5) == pytest.approx(1.0)

    def test_no_interp_mode(self):
        c = Curve()
        c.update_many([0.0, 1.0], [10.0, 20.0], reset=True)

        params = CurveParams()
        params.ip_kind = IP_KIND.NO_INTERP
        c.params = params

        # Direct lookup for existing times
        assert c.value(0.0) == pytest.approx(10.0)
        assert c.value(1.0) == pytest.approx(20.0)

        # Missing time returns NaN
        v = c.value(0.5)
        assert math.isnan(v)

    def test_remove_and_perturb(self):
        c = Curve()
        c.update_many([0, 1, 2], [10.0, 20.0, 30.0], reset=True)

        # Remove existing point
        assert c.remove(1) is True
        assert c.times == [0, 2]
        assert c.values == [10.0, 30.0]

        # Removing non-existing point
        assert c.remove(100) is False


class TestTwoFuncInterpolator:
    def test_requires_in_func_or_in_func_on_arrays(self):
        with pytest.raises(AssertionError):
            TwoFuncInterpolator(None, lambda t, v: v)

    def test_composed_interpolation(self):
        # in_func multiplies values by 2, out_func divides by 2 -- net effect is identity
        def in_func_on_arrays(xs, ys):
            return [2.0 * y for y in ys]

        def out_func(t, v):
            return v / 2.0

        tf = TwoFuncInterpolator(
            in_func=None,
            out_func=out_func,
            in_func_on_arrays=in_func_on_arrays,
        )

        xs = [0.0, 1.0]
        ys = [0.0, 10.0]

        ip = tf(xs, ys, kind='linear', fill_value='extrapolate', bounds_error=False)

        # At mid-point we should recover the simple linear interpolation of original ys
        assert ip(0.5) == pytest.approx(5.0)


class TestDateCurve:
    def test_dates_set_and_get(self):
        dc = DateCurve()

        d1 = date(2023, 1, 1)
        d2 = date(2023, 1, 5)
        dc.dates = [d1, d2]

        # Internal representation must be integer days from epoch
        assert isinstance(dc.times[0], int)
        assert dc.dates == [d1, d2]

    def test_update_and_value(self):
        dc = DateCurve()

        d1 = date(2023, 1, 1)
        d2 = date(2023, 1, 11)

        dc.update(d1, 0.0, reset=False)
        dc.update(d2, 10.0, reset=True)

        # Allow interpolation across the whole range
        dc.beginning_of_time = dc._to_number(d1)

        assert dc.value(d1) == pytest.approx(0.0)
        assert dc.value(d2) == pytest.approx(10.0)

        mid_date = d1 + (d2 - d1) / 2
        assert dc.value(mid_date) == pytest.approx(5.0)

    def test_beginning_of_time_helpers(self):
        dc = DateCurve()

        some_date = date(2020, 6, 15)
        num = dc._to_number(some_date)
        assert isinstance(num, int)
        assert dc._from_number(num) == some_date

        # beginning_of_time_set stores numeric value but exposes date via helper
        rc = dc.beginning_of_time_set('beginning_of_time', some_date)
        assert rc
        assert isinstance(dc.beginning_of_time, int)
        assert dc.beginning_of_time_as_date() == some_date
